from autolens.imaging import scaled_array
import numpy as np
import pytest
import os

test_data_dir = "{}/../test_files/array/".format(os.path.dirname(os.path.realpath(__file__)))

@pytest.fixture(name="array_grid")
def make_array_grid():
    return scaled_array.ScaledArray(np.zeros((5, 5)), pixel_scale=0.5)


class TestDataGrid(object):
    class TestConstructors(object):

        def test__constructor(self, array_grid):
            # Does the array grid class correctly instantiate as an instance of ndarray?
            assert array_grid.shape == (5, 5)
            assert array_grid.pixel_scale == 0.5
            assert isinstance(array_grid, np.ndarray)
            assert isinstance(array_grid, scaled_array.ScaledArray)

        def test__init__input_data_grid_single_value__all_attributes_correct_including_data_inheritance(
                self):
            data_grid = scaled_array.ScaledArray.single_value(value=5.0, shape=(3, 3),
                                                              pixel_scale=1.0)

            assert (data_grid == 5.0 * np.ones((3, 3))).all()
            assert data_grid.pixel_scale == 1.0
            assert data_grid.shape == (3, 3)
            assert data_grid.central_pixel_coordinates == (1.0, 1.0)
            assert data_grid.shape_arc_seconds == pytest.approx((3.0, 3.0))

        def test__init__input_data_grid_3x3__all_attributes_correct_including_data_inheritance(self):
            data_grid = scaled_array.ScaledArray(array=np.ones((3, 3)), pixel_scale=1.0)

            assert data_grid.pixel_scale == 1.0
            assert data_grid.shape == (3, 3)
            assert data_grid.central_pixel_coordinates == (1.0, 1.0)
            assert data_grid.shape_arc_seconds == pytest.approx((3.0, 3.0))
            assert (data_grid == np.ones((3, 3))).all()

        def test__init__input_data_grid_4x3__all_attributes_correct_including_data_inheritance(self):
            data_grid = scaled_array.ScaledArray(array=np.ones((4, 3)), pixel_scale=0.1)

            assert (data_grid == np.ones((4, 3))).all()
            assert data_grid.pixel_scale == 0.1
            assert data_grid.shape == (4, 3)
            assert data_grid.central_pixel_coordinates == (1.5, 1.0)
            assert data_grid.shape_arc_seconds == pytest.approx((0.4, 0.3))

        def test__from_fits__input_data_grid_3x3__all_attributes_correct_including_data_inheritance(self):
            data_grid = scaled_array.ScaledArray.from_fits(file_path=test_data_dir + '3x3_ones', hdu=0,
                                                           pixel_scale=1.0)

            assert (data_grid == np.ones((3, 3))).all()
            assert data_grid.pixel_scale == 1.0
            assert data_grid.shape == (3, 3)
            assert data_grid.central_pixel_coordinates == (1.0, 1.0)
            assert data_grid.shape_arc_seconds == pytest.approx((3.0, 3.0))

        def test__from_fits__input_data_grid_4x3__all_attributes_correct_including_data_inheritance(self):
            data_grid = scaled_array.ScaledArray.from_fits(file_path=test_data_dir + '4x3_ones', hdu=0,
                                                           pixel_scale=0.1)

            assert (data_grid == np.ones((4, 3))).all()
            assert data_grid.pixel_scale == 0.1
            assert data_grid.shape == (4, 3)
            assert data_grid.central_pixel_coordinates == (1.5, 1.0)
            assert data_grid.shape_arc_seconds == pytest.approx((0.4, 0.3))

    class TestCentralPixel:

        def test__3x3_grid__central_pixel_is_1_and_1(self):
            grid = scaled_array.ScaledArray(np.zeros((3, 3)), pixel_scale=0.1)
            assert grid.central_pixel_coordinates == (1, 1)

        def test__4x4_grid__central_pixel_is_1dot5_and_1dot5(self):
            grid = scaled_array.ScaledArray(np.zeros((4, 4)), pixel_scale=0.1)
            assert grid.central_pixel_coordinates == (1.5, 1.5)

        def test__5x3_grid__central_pixel_is_2_and_1(self):
            grid = scaled_array.ScaledArray(np.zeros((5, 3)), pixel_scale=0.1)
            assert grid.central_pixel_coordinates == (2, 1)

        def test__central_pixel_coordinates_5x5(self, array_grid):
            assert array_grid.central_pixel_coordinates == (2, 2)

    class TestConversion:

        def test__pixels_to_arcseconds(self, array_grid):
            assert array_grid.pixels_to_arc_seconds(1) == 0.5

        def test__arcseconds_to_pixels(self, array_grid):
            assert array_grid.arc_seconds_to_pixels(1) == 2

        def test__pixel_coordinates_to_arc_second_coordinates(self, array_grid):
            # Does the central pixel have (0, 0) coordinates in arcseconds?
            assert array_grid.pixel_coordinates_to_arc_second_coordinates((2, 2)) == (0, 0)

        def test__arc_second_coordinates_to_pixel_coordinates(self, array_grid):
            # Does the (0, 0) coordinates correspond to the central pixel?
            assert array_grid.arc_second_coordinates_to_pixel_coordinates((0, 0)) == (2, 2)

    class TestPad:

        def test__from_3x3_to_5x5(self):
            array = np.ones((3, 3))
            array[1, 1] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)
            modified = array.pad(new_dimensions=(5, 5))

            assert (modified == np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 1.0, 1.0, 1.0, 0.0],
                                          [0.0, 1.0, 2.0, 1.0, 0.0],
                                          [0.0, 1.0, 1.0, 1.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0]])).all()

            assert modified.shape == (5, 5)
            assert modified.shape_arc_seconds == (5.0, 5.0)

        def test__from_5x5_to_9x9(self):
            array = np.ones((5, 5))
            array[2, 2] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)
            modified = array.pad(new_dimensions=(9, 9))

            assert (modified == np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 1.0, 2.0, 1.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])).all()

            assert modified.shape == (9, 9)
            assert modified.shape_arc_seconds == (9.0, 9.0)

        def test__from_3x3_to_4x4__goes_to_5x5_to_keep_symmetry(self):
            array = np.ones((3, 3))
            array[1, 1] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)
            modified = array.pad(new_dimensions=(4, 4))

            assert (modified == np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 1.0, 1.0, 1.0, 0.0],
                                          [0.0, 1.0, 2.0, 1.0, 0.0],
                                          [0.0, 1.0, 1.0, 1.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0]])).all()

            assert modified.shape == (5, 5)
            assert modified.shape_arc_seconds == (5.0, 5.0)

        def test__from_5x5_to_8x8__goes_to_9x9_to_keep_symmetry(self):
            array = np.ones((5, 5))
            array[2, 2] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)
            modified = array.pad(new_dimensions=(8, 8))

            assert (modified == np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 1.0, 2.0, 1.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])).all()

            assert modified.shape == (9, 9)
            assert modified.shape_arc_seconds == (9.0, 9.0)

        def test__from_4x4_to_6x6(self):
            array = np.ones((4, 4))
            array[1:3, 1:3] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)
            modified = array.pad(new_dimensions=(6, 6))

            assert (modified == np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                                          [0.0, 1.0, 2.0, 2.0, 1.0, 0.0],
                                          [0.0, 1.0, 2.0, 2.0, 1.0, 0.0],
                                          [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])).all()

            assert modified.shape == (6, 6)
            assert modified.shape_arc_seconds == (6.0, 6.0)

        def test__from_4x4_to_8x8(self):
            array = np.ones((4, 4))
            array[1:3, 1:3] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)
            modified = array.pad(new_dimensions=(8, 8))

            assert (modified == np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 2.0, 2.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 2.0, 2.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])).all()

            assert modified.shape == (8, 8)
            assert modified.shape_arc_seconds == (8.0, 8.0)

        def test__from_4x4_to_5x5__goes_to_6x6_to_keep_symmetry(self):
            array = np.ones((4, 4))
            array[1:3, 1:3] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)
            modified = array.pad(new_dimensions=(5, 5))

            assert (modified == np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                                          [0.0, 1.0, 2.0, 2.0, 1.0, 0.0],
                                          [0.0, 1.0, 2.0, 2.0, 1.0, 0.0],
                                          [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])).all()

            assert modified.shape == (6, 6)
            assert modified.shape_arc_seconds == (6.0, 6.0)

        def test__from_4x4_to_7x7__goes_to_8x8_to_keep_symmetry(self):
            array = np.ones((4, 4))
            array[1:3, 1:3] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)
            modified = array.pad(new_dimensions=(7, 7))

            assert (modified == np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 2.0, 2.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 2.0, 2.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])).all()

            assert modified.shape == (8, 8)
            assert modified.shape_arc_seconds == (8.0, 8.0)

        def test__from_5x4_to_7x6(self):
            array = np.ones((5, 4))
            array[2, 1:3] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)
            modified = array.pad(new_dimensions=(7, 6))

            assert (modified == np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                                          [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                                          [0.0, 1.0, 2.0, 2.0, 1.0, 0.0],
                                          [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                                          [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])).all()

            assert modified.shape == (7, 6)
            assert modified.shape_arc_seconds == (7.0, 6.0)

        def test__from_2x3_to_6x7(self):
            array = np.ones((2, 3))
            array[0:2, 1] = 2.0
            array[1, 2] = 9

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)
            modified = array.pad(new_dimensions=(6, 7))

            assert (modified == np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 2.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 2.0, 9.0, 0.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])).all()

            assert modified.shape == (6, 7)
            assert modified.shape_arc_seconds == (6.0, 7.0)

        def test__from_2x3_to_5x6__goes_to_6x7_to_keep_symmetry(self):
            array = np.ones((2, 3))
            array[0:2, 1] = 2.0
            array[1, 2] = 9

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)
            modified = array.pad(new_dimensions=(5, 6))

            assert (modified == np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 2.0, 1.0, 0.0, 0.0],
                                          [0.0, 0.0, 1.0, 2.0, 9.0, 0.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])).all()

            assert modified.shape == (6, 7)
            assert modified.shape_arc_seconds == (6.0, 7.0)

        def test__x_size_smaller_than_array__raises_error(self):
            array = np.ones((5, 5))
            array[2, 2] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            with pytest.raises(ValueError):
                array.trim(new_dimensions=(3, 8))

        def test__y_size_smaller_than_array__raises_error(self):
            array = np.ones((5, 5))
            array[2, 2] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            with pytest.raises(ValueError):
                array.trim(new_dimensions=(8, 3))

        def test__pad_with_1s_instead(self):

            array = np.ones((3, 3))
            array[1, 1] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)
            modified = array.pad(new_dimensions=(5, 5), pad_value=1)

            assert (modified == np.array([[1.0, 1.0, 1.0, 1.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0, 1.0],
                                          [1.0, 1.0, 2.0, 1.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0, 1.0]])).all()

            assert modified.shape == (5, 5)
            assert modified.shape_arc_seconds == (5.0, 5.0)

    class TestTrim:

        def test__from_5x5_to_3x3(self):
            array = np.ones((5, 5))
            array[2, 2] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            modified = array.trim(new_dimensions=(3, 3))

            assert (modified == np.array([[1.0, 1.0, 1.0],
                                          [1.0, 2.0, 1.0],
                                          [1.0, 1.0, 1.0]])).all()

            assert modified.shape == (3, 3)
            assert modified.shape_arc_seconds == (3.0, 3.0)

        def test__from_7x7_to_3x3(self):
            array = np.ones((7, 7))
            array[3, 3] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            modified = array.trim(new_dimensions=(3, 3))

            assert (modified == np.array([[1.0, 1.0, 1.0],
                                          [1.0, 2.0, 1.0],
                                          [1.0, 1.0, 1.0]])).all()

            assert modified.shape == (3, 3)
            assert modified.shape_arc_seconds == (3.0, 3.0)

        def test__from_11x11_to_5x5(self):
            array = np.ones((11, 11))
            array[5, 5] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            modified = array.trim(new_dimensions=(5, 5))

            assert (modified == np.array([[1.0, 1.0, 1.0, 1.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0, 1.0],
                                          [1.0, 1.0, 2.0, 1.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0, 1.0]])).all()

            assert modified.shape == (5, 5)
            assert modified.shape_arc_seconds == (5.0, 5.0)

        def test__from_5x5_to_2x2__goes_to_3x3_to_keep_symmetry(self):
            array = np.ones((5, 5))
            array[2, 2] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            modified = array.trim(new_dimensions=(2, 2))

            assert (modified == np.array([[1.0, 1.0, 1.0],
                                          [1.0, 2.0, 1.0],
                                          [1.0, 1.0, 1.0]])).all()

            assert modified.shape == (3, 3)
            assert modified.shape_arc_seconds == (3.0, 3.0)

        def test__from_5x5_to_4x4__goes_to_5x5_to_keep_symmetry(self):
            array = np.ones((5, 5))
            array[2, 2] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            modified = array.trim(new_dimensions=(4, 4))

            assert (modified == np.array([[1.0, 1.0, 1.0, 1.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0, 1.0],
                                          [1.0, 1.0, 2.0, 1.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0, 1.0]])).all()

            assert modified.shape == (5, 5)
            assert modified.shape_arc_seconds == (5.0, 5.0)

        def test__from_11x11_to_4x4__goes_to_5x5_to_keep_symmetry(self):
            array = np.ones((11, 11))
            array[5, 5] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            modified = array.trim(new_dimensions=(4, 4))

            assert (modified == np.array([[1.0, 1.0, 1.0, 1.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0, 1.0],
                                          [1.0, 1.0, 2.0, 1.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0, 1.0]])).all()

            assert modified.shape == (5, 5)
            assert modified.shape_arc_seconds == (5.0, 5.0)

        def test__from_4x4_to_2x2(self):
            array = np.ones((4, 4))
            array[1:3, 1:3] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            modified = array.trim(new_dimensions=(2, 2))

            assert (modified == np.array([[2.0, 2.0],
                                          [2.0, 2.0]])).all()

            assert modified.shape == (2, 2)
            assert modified.shape_arc_seconds == (2.0, 2.0)

        def test__from_6x6_to_4x4(self):
            array = np.ones((6, 6))
            array[2:4, 2:4] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            modified = array.trim(new_dimensions=(4, 4))

            assert (modified == np.array([[1.0, 1.0, 1.0, 1.0],
                                          [1.0, 2.0, 2.0, 1.0],
                                          [1.0, 2.0, 2.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0]])).all()

            assert modified.shape == (4, 4)
            assert modified.shape_arc_seconds == (4.0, 4.0)

        def test__from_12x12_to_6x6(self):
            array = np.ones((12, 12))
            array[5:7, 5:7] = 2.0
            array[4, 4] = 9.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            modified = array.trim(new_dimensions=(6, 6))

            assert (modified == np.array([[1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                                          [1.0, 9.0, 1.0, 1.0, 1.0, 1.0],
                                          [1.0, 1.0, 2.0, 2.0, 1.0, 1.0],
                                          [1.0, 1.0, 2.0, 2.0, 1.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]])).all()

            assert modified.shape == (6, 6)
            assert modified.shape_arc_seconds == (6.0, 6.0)

        def test__from_4x4_to_3x3__goes_to_4x4_to_keep_symmetry(self):
            array = np.ones((4, 4))
            array[1:3, 1:3] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            modified = array.trim(new_dimensions=(3, 3))

            assert (modified == np.array([[1.0, 1.0, 1.0, 1.0],
                                          [1.0, 2.0, 2.0, 1.0],
                                          [1.0, 2.0, 2.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0]])).all()

            assert modified.shape == (4, 4)
            assert modified.shape_arc_seconds == (4.0, 4.0)

        def test__from_6x6_to_3x3_goes_to_4x4_to_keep_symmetry(self):
            array = np.ones((6, 6))
            array[2:4, 2:4] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            modified = array.trim(new_dimensions=(3, 3))

            assert (modified == np.array([[1.0, 1.0, 1.0, 1.0],
                                          [1.0, 2.0, 2.0, 1.0],
                                          [1.0, 2.0, 2.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0]])).all()

            assert modified.shape == (4, 4)
            assert modified.shape_arc_seconds == (4.0, 4.0)

        def test__from_12x12_to_5x5__goes_to_6x6_to_keep_symmetry(self):
            array = np.ones((12, 12))
            array[5:7, 5:7] = 2.0
            array[4, 4] = 9.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            modified = array.trim(new_dimensions=(5, 5))

            assert (modified == np.array([[1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                                          [1.0, 9.0, 1.0, 1.0, 1.0, 1.0],
                                          [1.0, 1.0, 2.0, 2.0, 1.0, 1.0],
                                          [1.0, 1.0, 2.0, 2.0, 1.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]])).all()

            assert modified.shape == (6, 6)
            assert modified.shape_arc_seconds == (6.0, 6.0)

        def test__from_5x4_to_3x2(self):
            array = np.ones((5, 4))
            array[2, 1:3] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            modified = array.trim(new_dimensions=(3, 2))

            assert (modified == np.array([[1.0, 1.0],
                                          [2.0, 2.0],
                                          [1.0, 1.0]])).all()

            assert modified.shape == (3, 2)
            assert modified.shape_arc_seconds == (3.0, 2.0)

        def test__from_4x5_to_2x3(self):
            array = np.ones((4, 5))
            array[1:3, 2] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            modified = array.trim(new_dimensions=(2, 3))

            assert (modified == np.array([[1.0, 2.0, 1.0],
                                          [1.0, 2.0, 1.0]])).all()

            assert modified.shape == (2, 3)
            assert modified.shape_arc_seconds == (2.0, 3.0)

        def test__from_5x4_to_4x3__goes_to_5x4_to_keep_symmetry(self):
            array = np.ones((5, 4))
            array[2, 1:3] = 2.0
            array[4, 3] = 9.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            modified = array.trim(new_dimensions=(4, 3))

            assert (modified == np.array([[1.0, 1.0, 1.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0],
                                          [1.0, 2.0, 2.0, 1.0],
                                          [1.0, 1.0, 1.0, 1.0],
                                          [1.0, 1.0, 1.0, 9.0]])).all()

            assert modified.shape == (5, 4)
            assert modified.shape_arc_seconds == (5.0, 4.0)

        def test__x_size_bigger_than_array__raises_error(self):
            array = np.ones((5, 5))
            array[2, 2] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            with pytest.raises(ValueError):
                array.trim(new_dimensions=(8, 3))

        def test__y_size_bigger_than_array__raises_error(self):
            array = np.ones((5, 5))
            array[2, 2] = 2.0

            array = scaled_array.ScaledArray(array, pixel_scale=1.0)

            with pytest.raises(ValueError):
                array.trim(new_dimensions=(3, 8))

    class TestGridCoordinates:

        def test__array_1x1__sets_up_arcsecond_coordinates(self):
            grid = scaled_array.ScaledArray(array=np.zeros((1, 1)), pixel_scale=1.0)

            grid_coordinates = grid.grid_coordinates

            assert (grid_coordinates == np.array([[[0.0, 0.0]]])).all()

            assert (grid_coordinates[0, 0] == np.array([[0.0, 0.0]])).all()

        def test__array_2x2__sets_up_arcsecond_coordinates(self):
            grid = scaled_array.ScaledArray(array=np.zeros((2, 2)), pixel_scale=1.0)

            grid_coordinates = grid.grid_coordinates

            assert (grid_coordinates == np.array([[[-0.5, -0.5], [-0.5, 0.5]],
                                                  [[0.5, -0.5], [0.5, 0.5]]])).all()

        def test__array_3x3__sets_up_arcsecond_coordinates(self):
            grid = scaled_array.ScaledArray(array=np.zeros((3, 3)), pixel_scale=1.0)

            grid_coordinates = grid.grid_coordinates
            print(list(grid_coordinates))

            assert (grid_coordinates == np.array([[[-1., -1.], [-1., 0.], [-1., 1.]],
                                                  [[0., -1.], [0., 0.], [0., 1.]],
                                                  [[1., -1.], [1., 0.], [1., 1.]]])).all()

        def test__array_4x4__sets_up_arcsecond_coordinates(self):
            grid = scaled_array.ScaledArray(array=np.zeros((4, 4)), pixel_scale=0.5)

            grid_coordinates = grid.grid_coordinates
            print(list(grid_coordinates))

            assert (grid_coordinates == np.array([[[-0.75, -0.75], [-0.75, -0.25], [-0.75, 0.25], [-0.75, 0.75]],
                                                  [[-0.25, -0.75], [-0.25, -0.25], [-0.25, 0.25], [-0.25, 0.75]],
                                                  [[0.25, -0.75], [0.25, -0.25], [0.25, 0.25], [0.25, 0.75]],
                                                  [[0.75, -0.75], [0.75, -0.25], [0.75, 0.25], [0.75, 0.75]]])).all()

        def test__array_2x3__sets_up_arcsecond_coordinates(self):
            grid = scaled_array.ScaledArray(array=np.zeros((2, 3)), pixel_scale=1.0)

            grid_coordinates = grid.grid_coordinates
            print(list(grid_coordinates))

            assert (grid_coordinates == np.array([[[-0.5, -1.], [-0.5, 0.], [-0.5, 1.]],
                                                  [[0.5, -1.], [0.5, 0.], [0.5, 1.]]])).all()

        def test__array_3x2__sets_up_arcsecond_coordinates(self):
            grid = scaled_array.ScaledArray(array=np.zeros((3, 2)), pixel_scale=1.0)

            grid_coordinates = grid.grid_coordinates

            assert (grid_coordinates == np.array([[[-1., -0.5], [-1., 0.5]],
                                                  [[0., -0.5], [0., 0.5]],
                                                  [[1., -0.5], [1., 0.5]]])).all()



import numpy as np

from autolens.profiles import light_profiles
from profiling import profiling_data
from profiling import tools
import numba


class Convolver(object):
    """Class to setup the 1D convolution of an image / mapping matrix.

    IMAGE FRAMES:
    ------------

    For a masked image, in 2D, one can compute for every unmasked pixel all other unmasked pixels it will blur light \
    into for a given PSF psf size, e.g.:

    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|     This is an example image.Mask, where:
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|     x = True (Pixel is masked and excluded from analysis)
    |x|x|x|o|o|o|x|x|x|x|     o = False (Pixel is not masked and included in analysis)
    |x|x|x|o|o|o|x|x|x|x|
    |x|x|x|o|o|o|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|

    Here, there are 9 unmasked pixels. Indexing of each unmasked pixel goes from the top-left corner right,
    and downwards, therefore:

    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|0|1|2|x|x|x|x|
    |x|x|x|3|4|5|x|x|x|x|
    |x|x|x|6|7|8|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|

    For every unmasked pixel, the Convolver over-lays the PSF psf and computes three quantities;

    image_frame_indexes - The indexes of all image pixels it will blur light into.
    image_frame_psfs - The psf values that overlap each image pixel it will blur light into.
    image_frame_length - The number of image-pixels it will blur light into (because unmasked pixels are excluded)

    For example, if we had the following 3x3 psf:

    |0.1|0.2|0.3|
    |0.4|0.5|0.6|
    |0.7|0.8|0.9|

    For pixel 0 above, when we overlap the psf 4 unmasked pixels overlap this psf, such that:

    image_frame_indexes = [0, 1, 3, 4]
    image_frame_psfs = [0.5, 0.6, 0.8, 0.9]
    image_frame_length = 4

    Noting that the other 5 psf values (0.1, 0.2, 0.3, 0.4, 0.7) overlap masked pixels and are thus discard.

    For pixel 1, we get the following results:

    image_frame_indexes = [0, 1, 2, 3, 4, 5]
    image_frame_psfs = [0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
    image_frame_lengths = 6

    In the majority of cases, there will be no unmasked pixels when the psf overlaps. This is the case above for \
    central pixel 4, where:

    image_frame_indexes = [0, 1, 2, 3, 4, 5, 6, 7, 8]
    image_frame_psfs = [0,1, 0.2, 0,3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
    image_frame_lengths = 6

    Once we have set up all these quantities, the convolution routine simply uses them to convolve a 1D array of image
    data / a mapping matrix image.

    BLURRING FRAMES:
    --------------

    Whilst the scheme above accounts for all blurred light within the mask, it does not account for the fact that
    pixels outside of the mask will also blur light into it. For galaxy light profiles, this effect is accounted for \
    using blurring frames, however it is omitted for mapping matrix images.

    First, a blurring mask is computed from a mask, which describes all pixels which are close enough to the mask \
    to blur light into it for a given psf size. Following the example above, the following blurring mask is \
    computed:

    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|     This is an example image.Mask, where:
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|o|o|o|o|o|x|x|x|     x = True (Pixel is masked and excluded from analysis)
    |x|x|o|x|x|x|o|x|x|x|     o = False (Pixel is not masked and included in analysis)
    |x|x|o|x|x|x|o|x|x|x|
    |x|x|o|x|x|x|o|x|x|x|
    |x|x|o|o|o|o|o|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|

    Indexing again goes from the top-left corner right and downwards:

    |x|x| x| x| x| x| x|x|x|x|
    |x|x| x| x| x| x| x|x|x|x|
    |x|x| x| x| x| x| x|x|x|x|
    |x|x| 0| 1| 2| 3| 4|x|x|x|
    |x|x| 5| x| x| x| 6|x|x|x|
    |x|x| 7| x| x| x| 8|x|x|x|
    |x|x| 9| x| x| x|10|x|x|x|
    |x|x|11|12|13|14|15|x|x|x|
    |x|x| x| x| x| x| x|x|x|x|
    |x|x| x| x| x| x| x|x|x|x|

    For every unmasked blurring-pixel, the Convolver over-lays the PSF psf and computes three quantities;

    blurring_frame_indexes - The indexes of all unmasked image pixels it will blur light into.
    bluring_frame_kernels - The psf values that overlap each unmasked image pixel it will blur light into.
    blurring_frame_length - The number of image-pixels it will blur light into (because unmasked pixels are excluded)

    Note, therefore, that the blurring frame does not perform any blurring which blurs light into other blurring pixels.
    It only performs computations which add light inside of the mask, which is the most computationally efficient method.

    For pixel 0 above, when we overlap the 3x3 psf above only 1 unmasked image pixels overlap the psf, such that:

    blurring_frame_indexes = [0]
    blurring_frame_psfs = [0.9]
    blurring_frame_length = 1

    For pixel 1 above, when we overlap the 3x3 psf above 2 unmasked image pixels overlap the psf, such that:

    blurring_frame_indexes = [0, 1]
    blurring_frame_psfs = [0.8, 0.9]
    blurring_frame_length = 2

    For pixel 3 above, when we overlap the 3x3 psf above 3 unmasked image pixels overlap the psf, such that:

    blurring_frame_indexes = [0, 1, 2]
    blurring_frame_psfs = [0.7, 0.8, 0.9]
    blurring_frame_length = 3
    """

    mask = None
    pixels_in_mask = None
    mask_index_array = None
    kernel = None
    kernel_shape = None
    kernel_max_size = None

    def setup_mask_index_array(self):
        count = 0
        for x in range(self.mask.shape[0]):
            for y in range(self.mask.shape[1]):
                if not self.mask[x, y]:
                    self.mask_index_array[x, y] = count
                    count += 1

    def setup_image_frames(self):
        image_index = 0
        self.image_frame_indexes = np.zeros((self.pixels_in_mask, self.kernel_max_size), dtype='int')
        self.image_frame_kernels = np.zeros((self.pixels_in_mask, self.kernel_max_size))
        self.image_frame_lengths = np.zeros((self.pixels_in_mask), dtype='int')
        for x in range(self.mask_index_array.shape[0]):
            for y in range(self.mask_index_array.shape[1]):
                if not self.mask[x][y]:
                    image_frame_indexes, image_frame_kernels = \
                        self.frame_at_coords_jit((x, y), self.mask, self.mask_index_array, self.kernel,
                                                 self.kernel_shape)

                    self.image_frame_indexes[image_index, :] = image_frame_indexes
                    self.image_frame_kernels[image_index, :] = image_frame_kernels
                    self.image_frame_lengths[image_index] = image_frame_indexes[image_frame_indexes >= 0].shape[0]
                    image_index += 1

    def frame_at_coords(self, coords):
        """
        Parameters
        ----------
        coords: (int, int)
            The image_grid of mask_index_array on which the frame should be centred
        psf_shape: (int, int)
            The shape of the psf for which this frame will be used
        Returns
        -------
        frame: ndarray
            A subset of mask_index_array of shape psf_shape where elements with image_grid outside of image_frame_indexes have
            value -1
        """
        half_x = int(self.kernel_shape[0] / 2)
        half_y = int(self.kernel_shape[1] / 2)

        frame = -1 * np.ones((self.kernel_max_size))
        kernel_frame = -1.0 * np.ones((self.kernel_max_size))

        count = 0
        for i in range(self.kernel_shape[0]):
            for j in range(self.kernel_shape[1]):
                x = coords[0] - half_x + i
                y = coords[1] - half_y + j
                if 0 <= x < self.mask_index_array.shape[0] and 0 <= y < self.mask_index_array.shape[1]:
                    value = self.mask_index_array[x, y]
                    if value >= 0 and not self.mask[x, y]:
                        frame[count] = value
                        kernel_frame[count] = self.kernel[i, j]
                        count += 1

        return frame, kernel_frame

    @staticmethod
    @numba.jit(nopython=True)
    def frame_at_coords_jit(coords, mask, mask_index_array, kernel, kernel_shape):
        """
        Parameters
        ----------
        coords: (int, int)
            The image_grid of mask_index_array on which the frame should be centred
        kernel_shape: (int, int)
            The shape of the psf for which this frame will be used
        Returns
        -------
        frame: ndarray
            A subset of mask_index_array of shape psf_shape where elements with image_grid outside of image_frame_indexes have
            value -1
        """

        kernel_max_size = kernel_shape[0] * kernel_shape[1]

        half_x = int(kernel_shape[0] / 2)
        half_y = int(kernel_shape[1] / 2)

        frame = -1 * np.ones((kernel_max_size))
        kernel_frame = -1.0 * np.ones((kernel_max_size))

        count = 0
        for i in range(kernel_shape[0]):
            for j in range(kernel_shape[1]):
                x = coords[0] - half_x + i
                y = coords[1] - half_y + j
                if 0 <= x < mask_index_array.shape[0] and 0 <= y < mask_index_array.shape[1]:
                    value = mask_index_array[x, y]
                    if value >= 0 and not mask[x, y]:
                        frame[count] = value
                        kernel_frame[count] = kernel[i, j]
                        count += 1

        return frame, kernel_frame


class ConvolverImage(Convolver):

    def __init__(self, mask, blurring_mask, kernel):
        """
        Class to create image frames and blurring frames used to convolve a psf with a 1D image of non-masked \
        values.

        Parameters
        ----------
        mask : Mask
            A mask where True eliminates data.
        burring_mask : Mask
            A mask of pixels outside the mask but whose light blurs into it after convolution.
        kernel : image.PSF or ndarray
            An array representing a PSF psf.
        """

        if kernel.shape[0] % 2 == 0 or kernel.shape[1] % 2 == 0:
            raise exc.KernelException("Kernel must be odd")
        if mask.shape != blurring_mask.shape:
            raise exc.KernelException("Mask and Blurring mask must be same shape to generate Convolver")

        self.mask = mask
        self.blurring_mask = blurring_mask
        self.mask_index_array = np.full(self.mask.shape, -1)
        self.pixels_in_mask = int(np.size(self.mask) - np.sum(self.mask))
        self.pixels_in_blurring_mask = int(np.size(blurring_mask) - np.sum(blurring_mask))

        self.setup_mask_index_array()

        self.kernel = kernel
        self.kernel_shape = kernel.shape
        self.kernel_max_size = self.kernel_shape[0] * self.kernel_shape[1]

        self.setup_image_frames()
        self.setup_blurring_frames()

    def setup_blurring_frames(self):

        image_index = 0
        self.blurring_frame_indexes = np.zeros((self.pixels_in_blurring_mask, self.kernel_max_size), dtype='int')
        self.blurring_frame_kernels = np.zeros((self.pixels_in_blurring_mask, self.kernel_max_size))
        self.blurring_frame_lengths = np.zeros((self.pixels_in_blurring_mask), dtype='int')
        for x in range(self.mask.shape[0]):
            for y in range(self.mask.shape[1]):
                if self.mask[x][y] and not self.blurring_mask[x, y]:
                    image_frame_indexes, image_frame_kernels = \
                        self.frame_at_coords_jit((x, y), self.mask, self.mask_index_array, self.kernel,
                                                 self.kernel_shape)

                    self.blurring_frame_indexes[image_index, :] = image_frame_indexes
                    self.blurring_frame_kernels[image_index, :] = image_frame_kernels
                    self.blurring_frame_lengths[image_index] = image_frame_indexes[image_frame_indexes >= 0].shape[0]
                    image_index += 1

    def convolve_image(self, image_array, blurring_array):
        """
        Parameters
        ----------
        blurring_array: [Float]
            An array representing the mapping of a source pixel to a set of image pixels within the blurring region.
        sub_shape: (int, int)
            Defines a sub_grid-region of the psf for which the result should be calculated
        image_array: [float]
            A 1D array
        Returns
        -------
        convolved_vector: [float]
            A vector convolved with the psf
        """

        new_array = np.zeros(image_array.shape)

        for image_index in range(len(image_array)):

            frame_indexes = self.image_frame_indexes[image_index]
            frame_kernels = self.image_frame_kernels[image_index]
            frame_length = self.image_frame_lengths[image_index]
            value = image_array[image_index]

            for kernel_index in range(frame_length):
                vector_index = frame_indexes[kernel_index]
                kernel = frame_kernels[kernel_index]
                new_array[vector_index] += value * kernel

        for image_index in range(len(blurring_array)):

            frame_indexes = self.blurring_frame_indexes[image_index]
            frame_kernels = self.blurring_frame_kernels[image_index]
            frame_length = self.blurring_frame_lengths[image_index]
            value = blurring_array[image_index]

            for kernel_index in range(frame_length):
                vector_index = frame_indexes[kernel_index]
                kernel = frame_kernels[kernel_index]
                new_array[vector_index] += value * kernel

        return new_array


sub_grid_size=4
psf_shape = (21, 21)
# psf_shape = (41, 41)
sersic = light_profiles.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0, intensity=0.1,
                                         effective_radius=0.8, sersic_index=4.0)

lsst = profiling_data.setup_class(name='LSST', pixel_scale=0.2, sub_grid_size=sub_grid_size, psf_shape=psf_shape)
euclid = profiling_data.setup_class(name='Euclid', pixel_scale=0.1, sub_grid_size=sub_grid_size, psf_shape=psf_shape)
hst = profiling_data.setup_class(name='HST', pixel_scale=0.05, sub_grid_size=sub_grid_size, psf_shape=psf_shape)
hst_up = profiling_data.setup_class(name='HSTup', pixel_scale=0.03, sub_grid_size=sub_grid_size, psf_shape=psf_shape)
ao = profiling_data.setup_class(name='AO', pixel_scale=0.01, sub_grid_size=sub_grid_size, psf_shape=psf_shape)

lsst_image = sersic.intensity_from_grid(grid=lsst.grids.image)
lsst_blurring_image = sersic.intensity_from_grid(grid=lsst.grids.blurring)
euclid_image = sersic.intensity_from_grid(grid=euclid.grids.image)
euclid_blurring_image = sersic.intensity_from_grid(grid=euclid.grids.blurring)
hst_image = sersic.intensity_from_grid(grid=hst.grids.image)
hst_blurring_image = sersic.intensity_from_grid(grid=hst.grids.blurring)
hst_up_image = sersic.intensity_from_grid(grid=hst_up.grids.image)
hst_up_blurring_image = sersic.intensity_from_grid(grid=hst_up.grids.blurring)
ao_image = sersic.intensity_from_grid(grid=ao.grids.image)
ao_blurring_image = sersic.intensity_from_grid(grid=ao.grids.blurring)

@tools.tick_toc_x1
def lsst_solution():
    lsst.masked_image.convolver_image.convolve_image_jit(image_array=lsst_image, blurring_array=lsst_blurring_image)

@tools.tick_toc_x1
def euclid_solution():
    euclid.masked_image.convolver_image.convolve_image_jit(image_array=euclid_image, blurring_array=euclid_blurring_image)

@tools.tick_toc_x1
def hst_solution():
    hst.masked_image.convolver_image.convolve_image_jit(image_array=hst_image, blurring_array=hst_blurring_image)

@tools.tick_toc_x1
def hst_up_solution():
    hst_up.masked_image.convolver_image.convolve_image_jit(image_array=hst_up_image, blurring_array=hst_up_blurring_image)

@tools.tick_toc_x1
def ao_solution():
    ao.masked_image.convolver_image.convolve_image_jit(image_array=ao_image, blurring_array=ao_blurring_image)

if __name__ == "__main__":
    lsst_solution()
    euclid_solution()
    hst_solution()
    hst_up_solution()
    ao_solution()


import numpy as np
import sklearn.cluster
import scipy.spatial

from autolens import exc
from autolens.pixelization import regularization
from autolens.pixelization import reconstruction
import numba

from autolens.imaging import mask


class PixelizationGrid(object):

    def __init__(self, shape=(3,3)):
        self.shape = shape

    def coordinate_grid_within_annulus(self, inner_radius, outer_radius, centre=(0., 0.)):

        x_pixel_scale = 2.0*outer_radius / self.shape[0]
        y_pixel_scale = 2.0*outer_radius / self.shape[1]

        central_pixel = float(self.shape[0] - 1) / 2, float(self.shape[1] - 1) / 2

        pixel_count = 0

        for x in range(self.shape[0]):
            for y in range(self.shape[1]):

                x_arcsec = ((x - central_pixel[0]) * x_pixel_scale) - centre[0]
                y_arcsec = ((y - central_pixel[1]) * y_pixel_scale) - centre[1]
                radius_arcsec = np.sqrt(x_arcsec ** 2 + y_arcsec ** 2)

                if radius_arcsec < outer_radius or radius_arcsec > inner_radius:

                    pixel_count += 1

        coordinates_array = np.zeros((pixel_count, 2))

        pixel_count = 0

        for x in range(self.shape[0]):
            for y in range(self.shape[1]):

                x_arcsec = ((x - central_pixel[0]) * x_pixel_scale) - centre[0]
                y_arcsec = ((y - central_pixel[1]) * y_pixel_scale) - centre[1]
                radius_arcsec = np.sqrt(x_arcsec ** 2 + y_arcsec ** 2)

                if radius_arcsec < outer_radius or radius_arcsec > inner_radius:

                    coordinates_array[pixel_count, 0] = x_arcsec
                    coordinates_array[pixel_count, 1] = y_arcsec

                    pixel_count += 1

        return coordinates_array





class Pixelization(object):

    def __init__(self, pixels=100, regularization_coefficients=(1.0,)):
        """
        Abstract base class for a pixelization, which discretizes a set of image and sub grid grid into \
        pixels. These pixels fit an image using a linear inversion, where a regularization matrix
        enforces smoothness between pixel values.

        A number of 1D and 2D arrays are used to represent mappings betwen image, sub, pix, and cluster pixels. The \
        nomenclature here follows grid_to_grid, such that it maps the index of a value on one grid to another. For \
        example:

        - pix_to_image[2] = 5 tells us that the 3rd pixelization-pixel maps to the 6th image-pixel.
        - sub_to_pix[4,2] = 2 tells us that the 5th sub-pixel maps to the 3rd pixelization-pixel.

        Parameters
        ----------
        pixels : int
            The number of pixels in the pixelization.
        regularization_coefficients : (float,)
            The regularization coefficients used to smooth the pix reconstruction.
        """
        self.pixels = pixels
        self.regularization_coefficients = regularization_coefficients

    @property
    def subscript(self):
        return 'pix'

    def mapping_matrix_from_sub_to_pix_jit(self, sub_to_pix, grids):
        return self.mapping_matrix_from_sub_to_pix_jitted(sub_to_pix, self.pixels, grids.image, grids.sub.sub_to_image,
                                                          grids.sub.sub_grid_fraction)

    @staticmethod
    @numba.jit(nopython=True)
    def mapping_matrix_from_sub_to_pix_jitted(sub_to_pix, pixels, grid_image, sub_to_image, sub_grid_fraction):

        mapping_matrix = np.zeros((grid_image.shape[0], pixels))

        for sub_index in range(sub_to_image.shape[0]):
            mapping_matrix[sub_to_image[sub_index], sub_to_pix[sub_index]] += sub_grid_fraction

        return mapping_matrix


class Rectangular(Pixelization):

    def __init__(self, shape=(3,3), regularization_coefficients=(1.0,)):
        """A rectangular pixelization where pixels appear on a Cartesian, uniform and rectangular grid \
        of  shape (rows, columns).

        Like an image grid, the indexing of the rectangular grid begins in the top-left corner and goes right and down.

        Parameters
        -----------
        shape : (int, int)
            The dimensions of the rectangular grid of pixels (x_pixels, y_pixel)
        regularization_coefficients : (float,)
            The regularization coefficients used to smooth the pix reconstruction.
        """

        if shape[0] <= 2 or shape[1] <= 2:
            raise exc.PixelizationException('The rectangular pixelization must be at least dimensions 3x3')

        super(Rectangular, self).__init__(shape[0] * shape[1], regularization_coefficients)

        self.shape = shape

    class Geometry(object):

        def __init__(self, x_min, x_max, x_pixel_scale, y_min, y_max, y_pixel_scale):
            """The geometry of a rectangular grid, defining where the grids top-left, top-right, bottom-left and \
            bottom-right corners are in arc seconds. The arc-second size of each rectangular pixel is also computed.

            Parameters
            -----------

            """
            self.x_min = x_min
            self.x_max = x_max
            self.x_pixel_scale = x_pixel_scale
            self.y_min = y_min
            self.y_max = y_max
            self.y_pixel_scale = y_pixel_scale

        def arc_second_to_pixel_index_x(self, coordinate):
            return np.floor((coordinate - self.x_min) / self.x_pixel_scale)

        def arc_second_to_pixel_index_y(self, coordinate):
            return np.floor((coordinate - self.y_min) / self.y_pixel_scale)

    def geometry_from_pix_sub_grid(self, pix_sub_grid, buffer=1e-8):
        """Determine the geometry of the rectangular grid, by alligning it with the outer-most pix_grid grid \
        plus a small buffer.

        Parameters
        -----------
        pix_sub_grid : [[float, float]]
            The x and y pix grid (or sub-coordinaates) which are to be matched with their pixels.
        buffer : float
            The size the grid-geometry is extended beyond the most exterior grid.
        """
        x_min = np.min(pix_sub_grid[:, 0]) - buffer
        x_max = np.max(pix_sub_grid[:, 0]) + buffer
        y_min = np.min(pix_sub_grid[:, 1]) - buffer
        y_max = np.max(pix_sub_grid[:, 1]) + buffer
        x_pixel_scale = (x_max - x_min) / self.shape[0]
        y_pixel_scale = (y_max - y_min) / self.shape[1]

        return self.Geometry(x_min, x_max, x_pixel_scale, y_min, y_max, y_pixel_scale)

    def neighbors_from_pixelization(self):
        """Compute the neighbors of every pixel as a list of the pixel index's each pixel shares a vertex with.

        The uniformity of the rectangular grid's geometry is used to compute this.
        """

        def compute_corner_neighbors(pix_neighbors):

            pix_neighbors[0] = [1, self.shape[1]]
            pix_neighbors[self.shape[1] - 1] = [self.shape[1] - 2, self.shape[1] + self.shape[1] - 1]
            pix_neighbors[self.pixels - self.shape[1]] = [self.pixels - self.shape[1] * 2,
                                                          self.pixels - self.shape[1] + 1]
            pix_neighbors[self.pixels - 1] = [self.pixels - self.shape[1] - 1, self.pixels - 2]

            return pix_neighbors

        def compute_top_edge_neighbors(pix_neighbors):

            for pix in range(1, self.shape[1] - 1):
                pixel_index = pix
                pix_neighbors[pixel_index] = [pixel_index - 1, pixel_index + 1, pixel_index + self.shape[1]]

            return pix_neighbors

        def compute_left_edge_neighbors(pix_neighbors):

            for pix in range(1, self.shape[0] - 1):
                pixel_index = pix * self.shape[1]
                pix_neighbors[pixel_index] = [pixel_index - self.shape[1], pixel_index + 1, pixel_index + self.shape[1]]

            return pix_neighbors

        def compute_right_edge_neighbors(pix_neighbors):

            for pix in range(1, self.shape[0] - 1):
                pixel_index = pix * self.shape[1] + self.shape[1] - 1
                pix_neighbors[pixel_index] = [pixel_index - self.shape[1], pixel_index - 1, pixel_index + self.shape[1]]

            return pix_neighbors

        def compute_bottom_edge_neighbors(pix_neighbors):

            for pix in range(1, self.shape[1] - 1):
                pixel_index = self.pixels - pix - 1
                pix_neighbors[pixel_index] = [pixel_index - self.shape[1], pixel_index - 1, pixel_index + 1]

            return pix_neighbors

        def compute_central_neighbors(pix_neighbors):

            for x in range(1, self.shape[0] - 1):
                for y in range(1, self.shape[1] - 1):
                    pixel_index = x * self.shape[1] + y
                    pix_neighbors[pixel_index] = [pixel_index - self.shape[1], pixel_index - 1, pixel_index + 1,
                                                  pixel_index + self.shape[1]]

            return pix_neighbors

        pixel_neighbors = [[] for _ in range(self.pixels)]

        pixel_neighbors = compute_corner_neighbors(pixel_neighbors)
        pixel_neighbors = compute_top_edge_neighbors(pixel_neighbors)
        pixel_neighbors = compute_left_edge_neighbors(pixel_neighbors)
        pixel_neighbors = compute_right_edge_neighbors(pixel_neighbors)
        pixel_neighbors = compute_bottom_edge_neighbors(pixel_neighbors)
        pixel_neighbors = compute_central_neighbors(pixel_neighbors)

        return pixel_neighbors

    def grid_to_pix_from_grid_jit(self, grid, geometry):
        """Compute the mappings between a set of image pixels (or sub-pixels) and pixels, using the image's
        traced pix-plane grid (or sub-grid) and the uniform rectangular pixelization's geometry.

        Parameters
        ----------
        grid : [[float, float]]
            The x and y pix grid (or sub-coordinates) which are to be matched with their pixels.
        geometry : Geometry
            The rectangular pixel grid's geometry.
        """
        return self.grid_to_pix_from_grid_jitted(grid, geometry.x_min, geometry.x_pixel_scale, geometry.y_min,
                                                 geometry.y_pixel_scale, self.shape[1]).astype(dtype='int')

    @staticmethod
    @numba.jit(nopython=True)
    def grid_to_pix_from_grid_jitted(grid, x_min, x_pixel_scale, y_min, y_pixel_scale, y_shape):

        grid_to_pix = np.zeros(grid.shape[0])

        for i in range(grid.shape[0]):

            x_pixel = np.floor((grid[i,0] - x_min) / x_pixel_scale)
            y_pixel = np.floor((grid[i,1] - y_min) / y_pixel_scale)

            grid_to_pix[i] = x_pixel * y_shape + y_pixel

        return grid_to_pix

    # TODO : RectangularRegWeight doesnt need sparse mask, but equivalent functions elsewhere do. Change to *kwrgs?

    def reconstructor_from_pix_grids(self, grids, borders, cluster_mask=None):
        """
        Compute the pixelization matrices of the rectangular pixelization by following these steps:

        1) Setup the rectangular grid geometry, by making its corner appear at the higher / lowest x and y pix sub-
        grid.
        2) Pair image and sub-image pixels to the rectangular grid using their traced grid and its geometry.

        Parameters
        ----------

        """
        relocated_grids = borders.relocated_grids_from_grids(grids)
        geometry = self.geometry_from_pix_sub_grid(relocated_grids.sub)
        pix_neighbors = self.neighbors_from_pixelization()
        image_to_pix = self.grid_to_pix_from_grid_jit(relocated_grids.image, geometry)
        sub_to_pix = self.grid_to_pix_from_grid_jit(relocated_grids.sub, geometry)

        mapping = self.mapping_matrix_from_sub_to_pix_jit(sub_to_pix, grids)
        regularization = self.regularization_matrix_from_pix_neighbors(pix_neighbors)

        return reconstruction.Reconstructor(mapping, regularization, image_to_pix, sub_to_pix)


class RectangularRegConst(Rectangular, regularization.RegularizationConstant):
    
    def __init__(self, shape=(3, 3), regularization_coefficients=(1.0,)):
        """A rectangular pixelization where pixels appear on a Cartesian, uniform and rectangular grid \
        of  shape (rows, columns).

        Like an image grid, the indexing of the rectangular grid begins in the top-left corner and goes right and down.

        Parameters
        -----------
        shape : (int, int)
            The dimensions of the rectangular grid of pixels (x_pixels, y_pixel)
        regularization_coefficients : (float,)
            The regularization coefficients used to smooth the pix reconstruction.
        """
        super(RectangularRegConst, self).__init__(shape, regularization_coefficients)

    @property
    def parameter_labels(self):
        return ['x', 'y', r'\lambda']

class RectangularRegWeighted(Rectangular, regularization.RegularizationWeighted):

    def __init__(self, shape=(3, 3), regularization_coefficients=(1.0, 1.0), pix_signal_scale=1.0):
        """A rectangular pixelization where pixels appear on a Cartesian, uniform and rectangular grid \
        of  shape (rows, columns).

        Like an image grid, the indexing of the rectangular grid begins in the top-left corner and goes right and down.

        Parameters
        -----------
        shape : (int, int)
            The dimensions of the rectangular grid of pixels (x_pixels, y_pixel)
        regularization_coefficients : (float,)
            The regularization coefficients used to smooth the pix reconstruction.
        """
        super(RectangularRegWeighted, self).__init__(shape, regularization_coefficients)
        self.pix_signal_scale = pix_signal_scale

    @property
    def parameter_labels(self):
        return ['x', 'y', r'\lambda1', r'\lambda2', 'V']

class Voronoi(Pixelization):

    def __init__(self, pixels=100, regularization_coefficients=(1.0,)):
        """
        Abstract base class for a Voronoi pixelization, which represents pixels as a set of centers where \
        all of the nearest-neighbor pix-grid (i.e. traced image-pixels) are mapped to them.

        This forms a Voronoi grid pix-plane, the properties of which are used for fast calculations, defining the \
        regularization matrix and visualization.

        Parameters
        ----------
        pixels : int
            The number of pixels in the pixelization.
        regularization_coefficients : (float,)
            The regularization coefficients used to smooth the pix reconstruction.
        """

        super(Voronoi, self).__init__(pixels, regularization_coefficients)

    @staticmethod
    def voronoi_from_cluster_grid(cluster_grid):
        """Compute the Voronoi grid of the pixelization, using the pixel centers.

        Parameters
        ----------
        cluster_grid : ndarray
            The x and y image_grid to derive the Voronoi grid_coords.
        """
        return scipy.spatial.Voronoi(cluster_grid, qhull_options='Qbb Qc Qx Qm')

    def neighbors_from_pixelization(self, ridge_points):
        """Compute the neighbors of every pixel as a list of the pixel index's each pixel \
        shares a vertex with.

        The ridge points of the Voronoi grid are used to derive this.

        Parameters
        ----------
        ridge_points : scipy.spatial.Voronoi.ridge_points
            Each Voronoi-ridge (two indexes representing a pixel mapping).
        """
        pix_neighbors = [[] for _ in range(self.pixels)]

        for pair in reversed(ridge_points):
            pix_neighbors[pair[0]].append(pair[1])
            pix_neighbors[pair[1]].append(pair[0])

        return pix_neighbors

    def image_to_pix_from_pixelization(self, grids, pix_centers, pix_neighbors, cluster_to_pix, cluster_mask):
        """ Compute the mappings between a set of image pixels and pixels, using the image's traced \
        pix-plane grid and the pixel centers.

        For the Voronoi pixelizations, a cluster set of 'cluster-pixels' are used to determine the pixelization. \
        These provide the mappings between only a sub-set of sub-pixels / image-pixels and pixels.

        To determine the complete set of sub-pixel to pixel mappings, we must therefore pair every sub-pixel to \
        its nearest pixel (using the sub-pixel's pix-plane coordinate and pixel center). Using a full \
        nearest neighbor search to do this is slow, thus the pixel neighbors (derived via the Voronoi grid) \
        is used to localize each nearest neighbor search.

        In this routine, some variables and function names refer to a 'cluster_pix_'. This term describes a \
        pixel that we have paired to a sub_coordinate using the cluster_coordinate of an image coordinate. \
        Thus, it may not actually be that sub_coordinate's closest pixel (the routine will eventually
        determine this).

        Parameters
        ----------

        cluster_mask
        grids
        pix_centers: [[float, float]]
            The coordinate of the center of every pixel.
        pix_neighbors : [[]]
            The neighboring pix_pixels of each pix_pixel, computed via the Voronoi grid_coords. \
            (e.g. if the fifth pix_pixel neighbors pix_pixels 7, 9 and 44, pix_neighbors[4] = [6, 8, 43])
        cluster_to_pix : [int]
            The mapping between every pixel and cluster-pixel (e.g. if the fifth pixel maps to \
            the 3rd cluster_pixel, cluster_to_pix[4] = 2).

        Returns
        ----------
        sub_to_pix : [int, int]
            The mapping between every sub-pixel and pixel. (e.g. if the fifth sub-pixel of the third \
            image-pixel maps to the 3rd pixel, sub_to_pix[2,4] = 2).

         """

        image_to_pix = np.zeros((grids.image.shape[0]), dtype=int)

        for image_index, pix_coordinate in enumerate(grids.image):
            nearest_cluster = cluster_mask.image_to_sparse[image_index]

            image_to_pix[image_index] = self.pair_image_and_pix(pix_coordinate, nearest_cluster,
                                                                pix_centers, pix_neighbors, cluster_to_pix)

        return image_to_pix

    def sub_to_pix_from_pixelization(self, grids, pix_centers, pix_neighbors, cluster_to_pix, cluster_mask):
        """ Compute the mappings between a set of sub-image pixels and pixels, using the image's traced \
        pix-plane sub-grid and the pixel centers. This uses the pix-neighbors to perform a graph \
        search when pairing pixels, for efficiency.

        For the Voronoi pixelizations, a cluster set of 'cluster-pixels' are used to determine the pixelization. \
        These provide the mappings between only a sub-set of sub-pixels / image-pixels and pixels.

        To determine the complete set of sub-pixel to pixel mappings, we must therefore pair every sub-pixel to \
        its nearest pixel (using the sub-pixel's pix-plane coordinate and pixel center). Using a full \
        nearest neighbor search to do this is slow, thus the pixel neighbors (derived via the Voronoi grid) \
        is used to localize each nearest neighbor search.

        In this routine, some variables and function names refer to a 'cluster_pix_'. This term describes a \
        pixel that we have paired to a sub_coordinate using the cluster_coordinate of an image coordinate. \
        Thus, it may not actually be that sub_coordinate's closest pixel (the routine will eventually
        determine this).

        Parameters
        ----------

        grids: mask.GridCollection
            A collection of coordinates for the masked image, subgrid and blurring grid
        cluster_mask: mask.SparseMask
            A mask describing the image pixels that should be used in pixel clustering
        pix_centers: [[float, float]]
            The coordinate of the center of every pixel.
        pix_neighbors : [[]]
            The neighboring pix_pixels of each pix_pixel, computed via the Voronoi grid_coords. \
            (e.g. if the fifth pix_pixel neighbors pix_pixels 7, 9 and 44, pix_neighbors[4] = [6, 8, 43])
        cluster_to_pix : [int]
            The mapping between every pixel and cluster-pixel (e.g. if the fifth pixel maps to \
            the 3rd cluster_pixel, pix_to_cluster[4] = 2).

        Returns
        ----------
        sub_to_pix : [int, int]
            The mapping between every sub-pixel and pixel. (e.g. if the fifth sub-pixel of the third \
            image-pixel maps to the 3rd pixel, sub_to_pix[2,4] = 2).

         """

        sub_to_pix = np.zeros((grids.sub.no_pixels,), dtype=int)

        for sub_index, sub_coordinate in enumerate(grids.sub):
            nearest_cluster = cluster_mask.image_to_sparse[grids.sub.sub_to_image[sub_index]]

            sub_to_pix[sub_index] = self.pair_image_and_pix(sub_coordinate, nearest_cluster, pix_centers,
                                                            pix_neighbors, cluster_to_pix)

        return sub_to_pix

    def pair_image_and_pix(self, coordinate, nearest_cluster, pix_centers, pix_neighbors, cluster_to_pix):
        """ Compute the mappings between a set of sub-image pixels and pixels, using the image's traced \
        pix-plane sub-grid and the pixel centers. This uses the pix-neighbors to perform a graph \
        search when pairing pixels, for efficiency.

        For the Voronoi pixelizations, a cluster set of 'cluster-pixels' are used to determine the pixelization. \
        These provide the mappings between only a sub-set of sub-pixels / image-pixels and pixels.

        To determine the complete set of sub-pixel to pixel mappings, we must therefore pair every sub-pixel to \
        its nearest pixel (using the sub-pixel's pix-plane coordinate and pixel center). Using a full \
        nearest neighbor search to do this is slow, thus the pixel neighbors (derived via the Voronoi grid) \
        is used to localize each nearest neighbor search.

        In this routine, some variables and function names refer to a 'cluster_pix_'. This term describes a \
        pixel that we have paired to a sub_coordinate using the cluster_coordinate of an image coordinate. \
        Thus, it may not actually be that sub_coordinate's closest pixel (the routine will eventually
        determine this).

        Parameters
        ----------
        coordinate : [float, float]
            The x and y pix sub-grid grid which are to be matched with their closest pixels.
        nearest_cluster : int
            The nearest pixel defined on the cluster-pixel grid.
        pix_centers: [[float, float]]
            The coordinate of the center of every pixel.
        pix_neighbors : [[]]
            The neighboring pix_pixels of each pix_pixel, computed via the Voronoi grid_coords. \
            (e.g. if the fifth pix_pixel neighbors pix_pixels 7, 9 and 44, pix_neighbors[4] = [6, 8, 43])
        cluster_to_pix : [int]
            The mapping between every cluster-pixel and pixel (e.g. if the fifth pixel maps to \
            the 3rd cluster_pixel, cluster_to_pix[4] = 2).
         """

        nearest_pix = cluster_to_pix[nearest_cluster]

        while True:

            pix_to_cluster_distance = self.distance_to_nearest_cluster_pix(coordinate, pix_centers, nearest_pix)

            neighboring_pix_index, sub_to_neighboring_pix_distance = \
                self.nearest_neighboring_pix_and_distance(coordinate, pix_centers,
                                                          pix_neighbors[nearest_pix])

            if pix_to_cluster_distance < sub_to_neighboring_pix_distance:
                return nearest_pix
            else:
                nearest_pix = neighboring_pix_index

    def distance_to_nearest_cluster_pix(self, coordinate, pix_centers, nearest_pix):
        nearest_cluster_pix_center = pix_centers[nearest_pix]
        return self.compute_squared_separation(coordinate, nearest_cluster_pix_center)

    def nearest_neighboring_pix_and_distance(self, coordinate, pix_centers, pix_neighbors):
        """For a given pix_pixel, we look over all its adjacent neighbors and find the neighbor whose distance is closest to
        our input coordinates.

        Parameters
        ----------
        coordinate : (float, float)
            The x and y coordinate to be matched with the neighboring set of pix_pixels.
        pix_centers: [(float, float)
            The pix_pixel centers the image_grid are matched with.
        pix_neighbors : []
            The neighboring pix_pixels of the cluster_grid pix_pixel the coordinate is currently matched with

        Returns
        ----------
        pix_neighbors_index : int
            The index in pix_pixel_centers of the closest pix_pixel neighbor.
        separation_from_neighbor : float
            The separation between the input coordinate and closest pix_pixel neighbor

        """

        separation_from_neighbor = list(map(lambda neighbors:
                                            self.compute_squared_separation(coordinate, pix_centers[neighbors]),
                                            pix_neighbors))

        closest_separation_index = min(range(len(separation_from_neighbor)),
                                       key=separation_from_neighbor.__getitem__)

        return pix_neighbors[closest_separation_index], separation_from_neighbor[closest_separation_index]

    @staticmethod
    def compute_squared_separation(coordinate1, coordinate2):
        """Computes the squared separation of two image_grid (no square root for efficiency)"""
        return (coordinate1[0] - coordinate2[0]) ** 2 + (coordinate1[1] - coordinate2[1]) ** 2


class Cluster(Voronoi):

    def __init__(self, pixels, regularization_coefficients=(1.0,)):
        """
        A cluster pixelization, which represents pixels as a set of centers where all of the nearest-neighbor \
        pix-grid (i.e. traced image-pixels) are mapped to them.

        For this pixelization, a set of cluster-pixels (defined in the image-plane as a cluster uniform grid of \
        image-pixels) determine the pixel centers .

        Parameters
        ----------
        pixels : int
            The number of pixels in the pixelization.
        regularization_coefficients : (float,)
            The regularization coefficients used to smooth the pix reconstruction.
        """
        super(Cluster, self).__init__(pixels, regularization_coefficients)

    def reconstructor_from_pix_grids(self, grids, borders, cluster_mask):
        """
        Compute the mapping matrix of the cluster pixelization by following these steps:

        1) Extract the cluster-grid (see grids.GridMapperCluster) from the pix-plane and use these as the \
        pixel centres.
        3) Derive a Voronoi grid using these pixel centres.
        4) Compute the mapping between all image sub-grid and pixels.
        5) Use these mappings to compute the mapping matrix.

        Parameters
        ----------
        grids: mask.GridCollection
            A collection of coordinates for the masked image, subgrid and blurring grid
        cluster_mask: mask.SparseMask
            A mask describing the image pixels that should be used in pixel clustering
        """

        relocated_grids = borders.relocated_grids_from_grids(grids)

        if self.pixels is not len(cluster_mask.sparse_to_image):
            raise exc.PixelizationException('ClusteringPixelization - The input number of pixels in the constructor'
                                            'is not the same as the length of the cluster_to_image mapper')

        pix_centers = grids.image[cluster_mask.sparse_to_image]
        cluster_to_pix = np.arange(0, self.pixels)
        voronoi = self.voronoi_from_cluster_grid(pix_centers)
        pix_neighbors = self.neighbors_from_pixelization(voronoi.ridge_points)
        image_to_pix = self.image_to_pix_from_pixelization(relocated_grids, pix_centers, pix_neighbors, cluster_to_pix,
                                                           cluster_mask)
        sub_to_pix = self.sub_to_pix_from_pixelization(relocated_grids, pix_centers, pix_neighbors, cluster_to_pix,
                                                       cluster_mask)

        mapping_matrix = self.mapping_matrix_from_sub_to_pix_jit(sub_to_pix, relocated_grids)
        regularization_matrix = self.regularization_matrix_from_pix_neighbors(pix_neighbors)

        return reconstruction.Reconstructor(mapping_matrix, regularization_matrix, image_to_pix, sub_to_pix)


class ClusterRegConst(Cluster, regularization.RegularizationConstant):

    def __init__(self, pixels, regularization_coefficients=(1.0,)):
        """
        A cluster pixelization, which represents pixels as a set of centers where all of the nearest-neighbor \
        pix-grid (i.e. traced image-pixels) are mapped to them.

        For this pixelization, a set of cluster-pixels (defined in the image-plane as a cluster uniform grid of \
        image-pixels) determine the pixel centers .

        Parameters
        ----------
        pixels : int
            The number of pixels in the pixelization.
        regularization_coefficients : (float,)
            The regularization coefficients used to smooth the pix reconstruction.
        """
        super(ClusterRegConst, self).__init__(pixels, regularization_coefficients)

    @property
    def parameter_labels(self):
        return ['N', r'\lambda1']

class ClusterRegWeighted(Cluster, regularization.RegularizationWeighted):

    def __init__(self, pixels, regularization_coefficients=(1.0, 1.0), pix_signal_scale=1.0):
        """
        A cluster pixelization, which represents pixels as a set of centers where all of the nearest-neighbor \
        pix-grid (i.e. traced image-pixels) are mapped to them.

        For this pixelization, a set of cluster-pixels (defined in the image-plane as a cluster uniform grid of \
        image-pixels) determine the pixel centers .

        Parameters
        ----------
        pixels : int
            The number of pixels in the pixelization.
        regularization_coefficients : (float,)
            The regularization coefficients used to smooth the pix reconstruction.
        """
        super(ClusterRegWeighted, self).__init__(pixels, regularization_coefficients)
        self.pix_signal_scale = pix_signal_scale

    @property
    def parameter_labels(self):
        return ['N', r'\lambda1', r'\lambda2', 'V']

class Amorphous(Voronoi):

    def __init__(self, pixels, regularization_coefficients=(1.0, 1.0, 2.0)):
        """
        An amorphous pixelization, which represents pixels as a set of centers where all of the \
        nearest-neighbor pix-grid (i.e. traced image-pixels) are mapped to them.

        For this pixelization, a set of cluster-pixels (defined in the image-plane as a cluster uniform grid of \
        image-pixels) are used to determine a set of pix-plane grid. These grid are then fed into a \
        weighted k-means clustering algorithm, such that the pixel centers adapt to the unlensed pix \
        surface-brightness profile.

        Parameters
        ----------
        pixels : int
            The number of pixels in the pixelization.
        regularization_coefficients : (float,)
            The regularization coefficients used to smooth the pix reconstruction.
        """
        super(Amorphous, self).__init__(pixels, regularization_coefficients)

    def kmeans_cluster(self, cluster_grid):
        """Perform k-means clustering on the cluster_grid to compute the k-means clusters which represent \
        pixels.

        Parameters
        ----------
        cluster_grid : ndarray
            The x and y cluster-grid which are used to derive the k-means pixelization.
        """
        kmeans = sklearn.cluster.KMeans(self.pixels)
        km = kmeans.fit(cluster_grid)
        return km.cluster_centers_, km.labels_

    def reconstructor_from_pix_grids(self, grids, borders, cluster_mask):
        """
        Compute the mapping matrix of the amorphous pixelization by following these steps:

        1) Extract the cluster-grid (see grids.GridMapperCluster) from the pix-plane.
        2) Performs weighted kmeans clustering on these cluster-grid to compute the pixel centres.
        3) Derive a Voronoi grid using these pixel centres.
        4) Compute the mapping between all image sub-grid and pixels.
        5) Use these mappings to compute the mapping matrix.

        Parameters
        ----------
        grids: mask.GridCollection
            A collection of coordinates for the masked image, subgrid and blurring grid
        cluster_mask: mask.SparseMask
            A mask describing the image pixels that should be used in pixel clustering
        """

        relocated_grids = borders.relocated_grids_from_grids(grids)

        cluster_grid = grids.image[cluster_mask.sparse_to_image]
        pix_centers, cluster_to_pix = self.kmeans_cluster(cluster_grid)
        voronoi = self.voronoi_from_cluster_grid(pix_centers)
        pix_neighbors = self.neighbors_from_pixelization(voronoi.ridge_points)
        image_to_pix = self.image_to_pix_from_pixelization(relocated_grids, pix_centers, pix_neighbors, cluster_to_pix,
                                                           cluster_mask)
        sub_to_pix = self.sub_to_pix_from_pixelization(relocated_grids, pix_centers, pix_neighbors, cluster_to_pix, cluster_mask)

        mapping_matrix = self.mapping_matrix_from_sub_to_pix_jit(sub_to_pix, relocated_grids)
        regularization_matrix = self.regularization_matrix_from_pix_neighbors(pix_neighbors)

        return reconstruction.Reconstructor(mapping_matrix, regularization_matrix, image_to_pix, sub_to_pix)


class AmorphousRegConst(Amorphous, regularization.RegularizationConstant):

    def __init__(self, pixels, regularization_coefficients=(1.0,)):
        """
        A cluster pixelization, which represents pixels as a set of centers where all of the nearest-neighbor \
        pix-grid (i.e. traced image-pixels) are mapped to them.

        For this pixelization, a set of cluster-pixels (defined in the image-plane as a cluster uniform grid of \
        image-pixels) determine the pixel centers .

        Parameters
        ----------
        pixels : int
            The number of pixels in the pixelization.
        regularization_coefficients : (float,)
            The regularization coefficients used to smooth the pix reconstruction.
        """
        super(AmorphousRegConst, self).__init__(pixels, regularization_coefficients)

    @property
    def parameter_labels(self):
        return ['N', r'\lambda1']

class AmorphousRegWeighted(Amorphous, regularization.RegularizationWeighted):

    def __init__(self, pixels, regularization_coefficients=(1.0, 1.0), pix_signal_scale=1.0):
        """
        A cluster pixelization, which represents pixels as a set of centers where all of the nearest-neighbor \
        pix-grid (i.e. traced image-pixels) are mapped to them.

        For this pixelization, a set of cluster-pixels (defined in the image-plane as a cluster uniform grid of \
        image-pixels) determine the pixel centers .

        Parameters
        ----------
        pixels : int
            The number of pixels in the pixelization.
        regularization_coefficients : (float,)
            The regularization coefficients used to smooth the pix reconstruction.
        """
        super(AmorphousRegWeighted, self).__init__(pixels, regularization_coefficients)
        self.pix_signal_scale = pix_signal_scale

    @property
    def parameter_labels(self):
        return ['N', r'\lambda1', r'\lambda2', 'V']


from profiling import profiling_data
from profiling import tools
from profiles import geometry_profiles
import numba
import numpy as np
import pytest

class SphericalProfile(geometry_profiles.Profile):

    def __init__(self, centre=(0.0, 0.0)):
        """ Generic elliptical profiles class to contain functions shared by light and mass profiles.

        Parameters
        ----------
        centre: (float, float)
            The coordinates of the centre of the profiles
        axis_ratio : float
            Ratio of profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        """
        super(SphericalProfile, self).__init__(centre)
        self.axis_ratio = 1.0
        self.phi = 0.0

    def transform_grid_to_reference_frame(self, grid):
        transformed = np.subtract(grid, self.centre)
        return transformed.view(geometry_profiles.TransformedGrid)

    def transform_grid_to_reference_frame_jitted(self, grid):
        return self.transform_grid_to_reference_jit(grid, self.centre[0], self.centre[1])

    @staticmethod
    @numba.jit(nopython=True)
    def transform_grid_to_reference_jit(grid, centre_x, centre_y):

        transformed = np.zeros(grid.shape)

        for i in range(grid.shape[0]):
            transformed[i,0] = np.subtract(grid[i,0], centre_x)
            transformed[i,1] = np.subtract(grid[i,1], centre_y)

        return transformed

sub_grid_size=4

lsst = profiling_data.setup_class(name='LSST', pixel_scale=0.2, sub_grid_size=sub_grid_size)
euclid = profiling_data.setup_class(name='Euclid', pixel_scale=0.1, sub_grid_size=sub_grid_size)
hst = profiling_data.setup_class(name='HST', pixel_scale=0.05, sub_grid_size=sub_grid_size)
hst_up = profiling_data.setup_class(name='HSTup', pixel_scale=0.03, sub_grid_size=sub_grid_size)
ao = profiling_data.setup_class(name='AO', pixel_scale=0.01, sub_grid_size=sub_grid_size)

geometry = SphericalProfile(centre=(0.0, 0.0))

assert (geometry.transform_grid_to_reference_frame(grid=lsst.coords.sub_grid_coords) ==
        pytest.approx(geometry.transform_grid_to_reference_frame_jitted(grid=lsst.coords.sub_grid_coords), 1e-4))

@tools.tick_toc_x20
def lsst_solution():
    geometry.transform_grid_to_reference_frame_jitted(grid=lsst.coords.sub_grid_coords)

@tools.tick_toc_x20
def euclid_solution():
    geometry.transform_grid_to_reference_frame_jitted(grid=euclid.coords.sub_grid_coords)

@tools.tick_toc_x20
def hst_solution():
    geometry.transform_grid_to_reference_frame_jitted(grid=hst.coords.sub_grid_coords)

@tools.tick_toc_x20
def hst_up_solution():
    geometry.transform_grid_to_reference_frame_jitted(grid=hst_up.coords.sub_grid_coords)

@tools.tick_toc_x20
def ao_solution():
    geometry.transform_grid_to_reference_frame_jitted(grid=ao.coords.sub_grid_coords)


if __name__ == "__main__":
    lsst_solution()
    euclid_solution()
    hst_solution()
    hst_up_solution()
    ao_solution()


from autolens.profiles import light_profiles
from profiling import profiling_data
import time
import pytest

subgrid_size=2

kernel_shape = (39,39)
sersic = light_profiles.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0, intensity=0.1,
                                         effective_radius=0.8, sersic_index=4.0)

lsst = profiling_data.setup_class(name='LSST', pixel_scale=0.2, subgrid_size=subgrid_size, psf_shape=kernel_shape)
lsst_image = sersic.intensity_from_grid(grid=lsst.coords.image_coords)
lsst_blurring_image = sersic.intensity_from_grid(grid=lsst.coords.blurring_coords)
lsst_kernel_convolver = lsst.masked_image.convolver.convolver_for_kernel(lsst.image.psf.trim(kernel_shape))

assert lsst_kernel_convolver.convolve_image_jit(image_array=lsst_image, blurring_array=lsst_blurring_image) == \
       pytest.approx(lsst_kernel_convolver.convolve_image_jit(image_array=lsst_image, blurring_array=lsst_blurring_image), 1e-4)

euclid = profiling_data.setup_class(name='Euclid', pixel_scale=0.1, subgrid_size=subgrid_size, psf_shape=kernel_shape)
euclid_image = sersic.intensity_from_grid(grid=euclid.coords.image_coords)
euclid_blurring_image = sersic.intensity_from_grid(grid=euclid.coords.blurring_coords)
euclid_kernel_convolver = euclid.masked_image.convolver.convolver_for_kernel(euclid.image.psf.trim(kernel_shape))

assert euclid_kernel_convolver.convolve_image_jit(image_array=euclid_image, blurring_array=euclid_blurring_image) == \
       pytest.approx(euclid_kernel_convolver.convolve_image_jit(image_array=euclid_image, blurring_array=euclid_blurring_image), 1e-4)

hst = profiling_data.setup_class(name='HST', pixel_scale=0.05, subgrid_size=subgrid_size, psf_shape=kernel_shape)
hst_image = sersic.intensity_from_grid(grid=hst.coords.image_coords)
hst_blurring_image = sersic.intensity_from_grid(grid=hst.coords.blurring_coords)
hst_kernel_convolver = hst.masked_image.convolver.convolver_for_kernel(hst.image.psf.trim(kernel_shape))

assert hst_kernel_convolver.convolve_image_jit(image_array=hst_image, blurring_array=hst_blurring_image) == \
       pytest.approx(hst_kernel_convolver.convolve_image_jit(image_array=hst_image, blurring_array=hst_blurring_image), 1e-4)

hst_up = profiling_data.setup_class(name='HSTup', pixel_scale=0.03, subgrid_size=subgrid_size, psf_shape=kernel_shape)
hst_up_image = sersic.intensity_from_grid(grid=hst_up.coords.image_coords)
hst_up_blurring_image = sersic.intensity_from_grid(grid=hst_up.coords.blurring_coords)
hst_up_kernel_convolver = hst_up.masked_image.convolver.convolver_for_kernel(hst_up.image.psf.trim(kernel_shape))

assert hst_up_kernel_convolver.convolve_image_jit(image_array=hst_up_image, blurring_array=hst_up_blurring_image) == \
       pytest.approx(hst_up_kernel_convolver.convolve_image_jit(image_array=hst_up_image, blurring_array=hst_up_blurring_image), 1e-4)

ao = profiling_data.setup_class(name='AO', pixel_scale=0.01, subgrid_size=subgrid_size, psf_shape=kernel_shape)
ao_image = sersic.intensity_from_grid(grid=ao.coords.image_coords)
ao_blurring_image = sersic.intensity_from_grid(grid=ao.coords.blurring_coords)
ao_kernel_convolver = ao.masked_image.convolver.convolver_for_kernel(ao.image.psf.trim(kernel_shape))

assert ao_kernel_convolver.convolve_image_jit(image_array=ao_image, blurring_array=ao_blurring_image) == \
       pytest.approx(ao_kernel_convolver.convolve_image_jit(image_array=ao_image, blurring_array=ao_blurring_image), 1e-4)

repeats = 1
def tick_toc(func):
    def wrapper():
        start = time.time()
        for _ in range(repeats):
            func()

        diff = time.time() - start
        print("{}: {}".format(func.__name__, diff))

    return wrapper

@tick_toc
def lsst_original_solution():
    lsst_kernel_convolver.convolve_image_jit(image_array=lsst_image, blurring_array=lsst_blurring_image)

@tick_toc
def lsst_jit_solution():
    lsst_kernel_convolver.convolve_image_jit(image_array=lsst_image, blurring_array=lsst_blurring_image)

@tick_toc
def euclid_original_solution():
    euclid_kernel_convolver.convolve_image_jit(image_array=euclid_image, blurring_array=euclid_blurring_image)

@tick_toc
def euclid_jit_solution():
    euclid_kernel_convolver.convolve_image_jit(image_array=euclid_image, blurring_array=euclid_blurring_image)

@tick_toc
def hst_original_solution():
    hst_kernel_convolver.convolve_image_jit(image_array=hst_image, blurring_array=hst_blurring_image)

@tick_toc
def hst_jit_solution():
    hst_kernel_convolver.convolve_image_jit(image_array=hst_image, blurring_array=hst_blurring_image)

@tick_toc
def hst_up_original_solution():
    hst_up_kernel_convolver.convolve_image_jit(image_array=hst_up_image, blurring_array=hst_up_blurring_image)

@tick_toc
def hst_up_jit_solution():
    hst_up_kernel_convolver.convolve_image_jit(image_array=hst_up_image, blurring_array=hst_up_blurring_image)

@tick_toc
def ao_original_solution():
    ao_kernel_convolver.convolve_image_jit(image_array=ao_image, blurring_array=ao_blurring_image)

@tick_toc
def ao_jit_solution():
    ao_kernel_convolver.convolve_image_jit(image_array=ao_image, blurring_array=ao_blurring_image)

# @tick_toc
# def jitted_solution():
#     kernel_convolver.convolve_array_jit(data)

if __name__ == "__main__":
    lsst_original_solution()
    lsst_jit_solution()

    print()

    euclid_original_solution()
    euclid_jit_solution()

    print()

    hst_original_solution()
    hst_jit_solution()
    #
    print()

    hst_up_original_solution()
    hst_up_jit_solution()

    print()

    ao_original_solution()
    ao_jit_solution()


from autolens.pipeline import phase as ph
from autolens.autopipe import non_linear as nl
from autolens.analysis import galaxy_prior as gp
from autolens.pixelization import pixelization as px
from autolens.imaging import mask as msk
from autolens.profiles import light_profiles, mass_profiles

import logging

logger = logging.getLogger(__name__)


class Pipeline(object):
    def __init__(self, *phases):
        """

        Parameters
        ----------
        phases: [ph.Phase]
            Phases
        """
        self.phases = phases

    def run(self, image):
        results = []
        for i, phase in enumerate(self.phases):
            logger.info("Running Phase {} (Number {})".format(phase.name, i))
            results.append(phase.run(image, ph.ResultsCollection(results)))
        return results

    def __add__(self, other):
        """
        Compose two pipelines

        Parameters
        ----------
        other: Pipeline
            Another pipeline

        Returns
        -------
        composed_pipeline: Pipeline
            A pipeline that runs all the  phases from this pipeline and then all the phases from the other pipeline
        """
        return Pipeline(*(self.phases + other.phases))


def make_source_only_pipeline():
    # 1) Mass: SIE+Shear
    #    Source: Sersic
    #    NLO: LM

    def mask_function(img):
        return msk.Mask.circular(img.shape_arc_seconds, img.pixel_scale, 2)

    phase1 = ph.SourceLensPhase(
        lens_galaxy=gp.GalaxyPrior(
            sie=mass_profiles.SphericalIsothermal,
            shear=mass_profiles.ExternalShear),
        source_galaxy=gp.GalaxyPrior(
            sersic=light_profiles.EllipticalSersic),
        mask_function=mask_function)

    # 2) Mass: SIE+Shear (priors from phase 1)
    #    Source: 'smooth' pixelization (include regularization parameter(s) in the model)
    #    NLO: LM

    class PriorLensPhase(ph.PixelizedSourceLensPhase):
        def pass_priors(self, previous_results):
            self.lens_galaxy = previous_results.last.variable.lens_galaxy

    phase2 = PriorLensPhase(pixelization=px.RectangularRegConst,
                            mask_function=mask_function)

    class ConstantLensPhase(ph.PixelizedSourceLensPhase):
        def pass_priors(self, previous_results):
            self.lens_galaxy = previous_results.last.constant.lens_galaxy

    phase3 = ConstantLensPhase(pixelization=px.RectangularRegConst,
                               mask_function=mask_function)

    return Pipeline(phase1, phase2, phase3)


def make_profile_pipeline(name="profile_pipeline", optimizer_class=nl.MultiNest):
    # 1) Lens Light : EllipticalSersic
    #    Mass: None
    #    Source: None
    #    NLO: MultiNest
    #    Image : Observed Image
    #    Mask : Circle - 3.0"

    phase1 = ph.LensOnlyPhase(lens_galaxy=gp.GalaxyPrior(elliptical_sersic=light_profiles.EllipticalSersic),
                              optimizer_class=optimizer_class, name="{}/phase1".format(name))

    class LensSubtractedPhase(ph.SourceLensPhase):
        def modify_image(self, masked_image, previous_results):
            return masked_image - previous_results.last.lens_galaxy_image

        def pass_priors(self, previous_results):
            self.lens_galaxy.sie.centre = previous_results.last.variable.lens_galaxy.elliptical_sersic.centre

    # 2) Lens Light : None
    #    Mass: SIE (use lens light profile centre from previous phase as prior on mass profile centre)
    #    Source: EllipticalSersic
    #    NLO: MultiNest
    #    Image : Lens Subtracted Image (previous phase)
    #    Mask : Annulus (0.4" - 3.0")

    def annular_mask_function(img):
        return msk.Mask.annular(img.shape_arc_seconds, pixel_scale=img.pixel_scale, inner_radius=0.4,
                                outer_radius=3.)

    phase2 = LensSubtractedPhase(lens_galaxy=gp.GalaxyPrior(sie=mass_profiles.SphericalIsothermal),
                                 source_galaxy=gp.GalaxyPrior(elliptical_sersic=light_profiles.EllipticalSersic),
                                 optimizer_class=optimizer_class,
                                 mask_function=annular_mask_function,
                                 name="{}/phase2".format(name))

    # 3) Lens Light : Elliptical Sersic (Priors phase 1)
    #    Mass: SIE (Priors phase 2)
    #    Source : Elliptical Sersic (Priors phase 2)
    #    NLO : MultiNest
    #    Image : Observed Image
    #    Mask : Circle - 3.0"

    class CombinedPhase(ph.SourceLensPhase):
        def pass_priors(self, previous_results):
            self.lens_galaxy = gp.GalaxyPrior(
                elliptical_sersic=previous_results.first.variable.lens_galaxy.elliptical_sersic,
                sie=previous_results.last.variable.lens_galaxy.sie)
            self.source_galaxy = previous_results.last.variable.source_galaxy

    phase3 = CombinedPhase(optimizer_class=optimizer_class,
                           name="{}/phase3".format(name))

    # 3H) Hyper-Parameters: Make Lens Galaxy and Source Galaxy Hyper-Galaxies.
    #     Lens Light / Mass / Source - Fix parameters to phase 3 most likely result
    #     NLO : DownhillSimplex
    #     Image : Observed Image
    #     Mask : Circle - 3.0"

    phase3h = ph.SourceLensHyperGalaxyPhase(name="{}/phase3h".format(name))

    # 4) Repeat phase 3, using its priors and the hyper-galaxies fixed to their optimized values.
    #    Lens Light : Elliptical Sersic (Priors phase 3)
    #    Mass: SIE (Priors phase 3)
    #    Source : Elliptical Sersic (Priors phase 3)
    #    NLO : MultiNest
    #    Image : Observed Image
    #    Mask : Circle - 3.0"

    class CombinedPhase2(ph.SourceLensPhase):
        def pass_priors(self, previous_results):
            phase_3_results = previous_results[2]
            self.lens_galaxy = phase_3_results.variable.lens_galaxy
            self.source_galaxy = phase_3_results.variable.source_galaxy
            self.lens_galaxy.hyper_galaxy = previous_results.last.constant.lens_galaxy.hyper_galaxy
            self.source_galaxy.hyper_galaxy = previous_results.last.constant.source_galaxy.hyper_galaxy

    phase4 = CombinedPhase2(optimizer_class=optimizer_class, name="{}/phase4".format(name))

    return Pipeline(phase1, phase2, phase3, phase3h, phase4)



import numpy as np
from autolens import exc
import numba

"""
This module is for the application of convolution to sparse_grid vectors.

Take a simple mask:

[[True, False, True],
 [False, False, False],
 [True, False, True]]
 
Here True means that the value is masked.

A set of values in a corresponding image_grid might be represented in a 1D array:

[2, 8, 2, 5, 7, 5, 3, 1, 4]

This module allows us to find the relationships between data_to_image in a mask for a psf of a given sub_grid_size so that
convolutions can be efficiently applied to reduced arrays such as the one above.

A Convolver can be created for a given mask:

frame_maker = Convolver(self.mask)

This can then produce a convolver_image for any given psf shape and corresponding blurring region mask:

convolver_image = frame_maker.convolver_for_kernel_shape((3, 3), blurring_region_mask)

Here the blurring region mask describes the region under the mask from which a given PSF psf may blur pixels. If the
regular mask specifies True for a givena pixel and the blurring region mask False then that pixel will be blurred in
using the blurring array.

A convolver_image can then be made for any given psf:

kernel_convolver = convolver_image.convolver_for_kernel(psf)

Which is applied to a reduced array and blurring array:

convolved_array = convolver_image.convolve_array(array, blurring_array)

The array is pixels within the non-masked region, whilst the blurring array is pixels outside of the non-masked region
but inside of the blurring mask region.

The convolver_image can also be applied for some sub_grid-shape of the psf:

convolved_vector = convolver_image.convolve_vector(vector, sub_shape=(3, 3))

Or applied to a whole mapping matrix:

convolved_mapping_matrix = convolver_image.convolve_mapping_matrix(mapping_matrix)

Where the mapping matrix is an array of dictionaries with each index of the array corresponding to a source pixel.

It is also possible to specify a blurring region mask:

Convolver(self.mask, blurring_region_mask)

PSF will be calculated from values that are masked by mask but not by blurring region mask. That is, any entry with a 
True value for mask and a False value for blurring region mask may contribute to the PSF convolved value of a nearby
entry with a False value for mask.

"""


class Convolver(object):
    """Class to setup the 1D convolution of an image / mapping matrix.

    IMAGE FRAMES:
    ------------

    For a masked image, in 2D, one can compute for every unmasked pixel all other unmasked pixels it will blur light \
    into for a given PSF psf size, e.g.:

    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|     This is an example image.Mask, where:
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|     x = True (Pixel is masked and excluded from analysis)
    |x|x|x|o|o|o|x|x|x|x|     o = False (Pixel is not masked and included in analysis)
    |x|x|x|o|o|o|x|x|x|x|
    |x|x|x|o|o|o|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|

    Here, there are 9 unmasked pixels. Indexing of each unmasked pixel goes from the top-left corner right,
    and downwards, therefore:

    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|0|1|2|x|x|x|x|
    |x|x|x|3|4|5|x|x|x|x|
    |x|x|x|6|7|8|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|

    For every unmasked pixel, the Convolver over-lays the PSF psf and computes three quantities;

    image_frame_indexes - The indexes of all image pixels it will blur light into.
    image_frame_psfs - The psf values that overlap each image pixel it will blur light into.
    image_frame_length - The number of image-pixels it will blur light into (because unmasked pixels are excluded)

    For example, if we had the following 3x3 psf:

    |0.1|0.2|0.3|
    |0.4|0.5|0.6|
    |0.7|0.8|0.9|

    For pixel 0 above, when we overlap the psf 4 unmasked pixels overlap this psf, such that:

    image_frame_indexes = [0, 1, 3, 4]
    image_frame_psfs = [0.5, 0.6, 0.8, 0.9]
    image_frame_length = 4

    Noting that the other 5 psf values (0.1, 0.2, 0.3, 0.4, 0.7) overlap masked pixels and are thus discard.

    For pixel 1, we get the following results:

    image_frame_indexes = [0, 1, 2, 3, 4, 5]
    image_frame_psfs = [0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
    image_frame_lengths = 6

    In the majority of cases, there will be no unmasked pixels when the psf overlaps. This is the case above for \
    central pixel 4, where:

    image_frame_indexes = [0, 1, 2, 3, 4, 5, 6, 7, 8]
    image_frame_psfs = [0,1, 0.2, 0,3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
    image_frame_lengths = 6

    Once we have set up all these quantities, the convolution routine simply uses them to convolve a 1D array of image
    data / a mapping matrix image.

    BLURRING FRAMES:
    --------------

    Whilst the scheme above accounts for all blurred light within the mask, it does not account for the fact that
    pixels outside of the mask will also blur light into it. For galaxy light profiles, this effect is accounted for \
    using blurring frames, however it is omitted for mapping matrix images.

    First, a blurring mask is computed from a mask, which describes all pixels which are close enough to the mask \
    to blur light into it for a given psf size. Following the example above, the following blurring mask is \
    computed:

    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|     This is an example image.Mask, where:
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|o|o|o|o|o|x|x|x|     x = True (Pixel is masked and excluded from analysis)
    |x|x|o|x|x|x|o|x|x|x|     o = False (Pixel is not masked and included in analysis)
    |x|x|o|x|x|x|o|x|x|x|
    |x|x|o|x|x|x|o|x|x|x|
    |x|x|o|o|o|o|o|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|

    Indexing again goes from the top-left corner right and downwards:

    |x|x| x| x| x| x| x|x|x|x|
    |x|x| x| x| x| x| x|x|x|x|
    |x|x| x| x| x| x| x|x|x|x|
    |x|x| 0| 1| 2| 3| 4|x|x|x|
    |x|x| 5| x| x| x| 6|x|x|x|
    |x|x| 7| x| x| x| 8|x|x|x|
    |x|x| 9| x| x| x|10|x|x|x|
    |x|x|11|12|13|14|15|x|x|x|
    |x|x| x| x| x| x| x|x|x|x|
    |x|x| x| x| x| x| x|x|x|x|

    For every unmasked blurring-pixel, the Convolver over-lays the PSF psf and computes three quantities;

    blurring_frame_indexes - The indexes of all unmasked image pixels it will blur light into.
    bluring_frame_kernels - The psf values that overlap each unmasked image pixel it will blur light into.
    blurring_frame_length - The number of image-pixels it will blur light into (because unmasked pixels are excluded)

    Note, therefore, that the blurring frame does not perform any blurring which blurs light into other blurring pixels.
    It only performs computations which add light inside of the mask, which is the most computationally efficient method.

    For pixel 0 above, when we overlap the 3x3 psf above only 1 unmasked image pixels overlap the psf, such that:

    blurring_frame_indexes = [0]
    blurring_frame_psfs = [0.9]
    blurring_frame_length = 1

    For pixel 1 above, when we overlap the 3x3 psf above 2 unmasked image pixels overlap the psf, such that:

    blurring_frame_indexes = [0, 1]
    blurring_frame_psfs = [0.8, 0.9]
    blurring_frame_length = 2

    For pixel 3 above, when we overlap the 3x3 psf above 3 unmasked image pixels overlap the psf, such that:

    blurring_frame_indexes = [0, 1, 2]
    blurring_frame_psfs = [0.7, 0.8, 0.9]
    blurring_frame_length = 3
    """

    def __init__(self, mask, psf):
        """
        Class to create image frames and blurring frames used to convolve a psf with a 1D image of non-masked \
        values.

        Parameters
        ----------
        mask : Mask
            A mask where True eliminates data.
        burring_mask : Mask
            A mask of pixels outside the mask but whose light blurs into it after convolution.
        psf : image.PSF or ndarray
            An array representing a PSF psf.
        """

        if psf.shape[0] % 2 == 0 or psf.shape[1] % 2 == 0:
            raise exc.KernelException("PSF kernel must be odd")

        self.mask_index_array = np.full(mask.shape, -1)
        self.pixels_in_mask = int(np.size(mask) - np.sum(mask))

        count = 0
        for x in range(mask.shape[0]):
            for y in range(mask.shape[1]):
                if not mask[x, y]:
                    self.mask_index_array[x, y] = count
                    count += 1

        self.psf = psf
        self.psf_shape = psf.shape
        self.psf_max_size = self.psf_shape[0] * self.psf_shape[1]

        image_index = 0
        self.image_frame_indexes = np.zeros((self.pixels_in_mask, self.psf_max_size), dtype='int')
        self.image_frame_psfs = np.zeros((self.pixels_in_mask, self.psf_max_size))
        self.image_frame_lengths = np.zeros((self.pixels_in_mask), dtype='int')
        for x in range(self.mask_index_array.shape[0]):
            for y in range(self.mask_index_array.shape[1]):
                if not mask[x][y]:
                    image_frame_indexes, image_frame_psfs = self.frame_at_coords_jit((x, y), mask,
                                                                                        self.mask_index_array, self.psf[:, :])
                    self.image_frame_indexes[image_index, :] = image_frame_indexes
                    self.image_frame_psfs[image_index, :] = image_frame_psfs
                    self.image_frame_lengths[image_index] = image_frame_indexes[image_frame_indexes >= 0].shape[0]
                    image_index += 1

    @staticmethod
    @numba.jit(nopython=True)
    def frame_at_coords_jit(coords, mask, mask_index_array, psf):
        """
        Parameters
        ----------
        coords: (int, int)
            The image_grid of mask_index_array on which the frame should be centred
        psf_shape: (int, int)
            The shape of the psf for which this frame will be used
        Returns
        -------
        frame: ndarray
            A subset of mask_index_array of shape psf_shape where elements with image_grid outside of image_frame_indexes have
            value -1
        """

        psf_shape = psf.shape
        psf_max_size = psf_shape[0] * psf_shape[1]

        half_x = int(psf_shape[0] / 2)
        half_y = int(psf_shape[1] / 2)

        frame = -1*np.ones((psf_max_size))
        psf_frame = -1.0*np.ones((psf_max_size))

        count = 0
        for i in range(psf_shape[0]):
            for j in range(psf_shape[1]):
                x = coords[0] - half_x + i
                y = coords[1] - half_y + j
                if 0 <= x < mask_index_array.shape[0] and 0 <= y < mask_index_array.shape[1]:
                    value = mask_index_array[x, y]
                    if value >= 0 and not mask[x, y]:
                        frame[count] = value
                        psf_frame[count] = psf[i, j]
                        count += 1

        return frame, psf_frame


class ConvolverImage(Convolver):

    def __init__(self, mask, blurring_mask, psf):
        """
        Class to create image frames and blurring frames used to convolve a psf with a 1D image of non-masked \
        values.

        Parameters
        ----------
        mask : Mask
            A mask where True eliminates data.
        burring_mask : Mask
            A mask of pixels outside the mask but whose light blurs into it after convolution.
        psf : image.PSF or ndarray
            An array representing a PSF psf.
        """

        if mask.shape != blurring_mask.shape:
            raise exc.KernelException("Mask and Blurring mask must be same shape to generate Convolver")

        super(ConvolverImage, self).__init__(mask, psf)

        blurring_mask = blurring_mask
        self.pixels_in_blurring_mask = int(np.size(blurring_mask) - np.sum(blurring_mask))

        image_index = 0
        self.blurring_frame_indexes = np.zeros((self.pixels_in_blurring_mask, self.psf_max_size), dtype='int')
        self.blurring_frame_psfs = np.zeros((self.pixels_in_blurring_mask, self.psf_max_size))
        self.blurring_frame_lengths = np.zeros((self.pixels_in_blurring_mask), dtype='int')
        for x in range(mask.shape[0]):
            for y in range(mask.shape[1]):
                if mask[x][y] and not blurring_mask[x, y]:
                    image_frame_indexes, image_frame_psfs = self.frame_at_coords_jit((x, y), mask,
                                                                                        self.mask_index_array, self.psf)
                    self.blurring_frame_indexes[image_index, :] = image_frame_indexes
                    self.blurring_frame_psfs[image_index, :] = image_frame_psfs
                    self.blurring_frame_lengths[image_index] = image_frame_indexes[image_frame_indexes >= 0].shape[0]
                    image_index += 1

    def convolve_image_jit(self, image_array, blurring_array):
        return self.convolve_image_jitted(image_array, self.image_frame_indexes,
                                          self.image_frame_psfs, self.image_frame_lengths,
                                          blurring_array, self.blurring_frame_indexes,
                                          self.blurring_frame_psfs, self.blurring_frame_lengths)

    @staticmethod
    @numba.jit(nopython=True)
    def convolve_image_jitted(image_array, image_frame_indexes, image_frame_kernels, image_frame_lengths,
                              blurring_array, blurring_frame_indexes, blurring_frame_kernels, blurring_frame_lengths):

        new_array = np.zeros(image_array.shape)

        for image_index in range(len(image_array)):

            frame_indexes = image_frame_indexes[image_index]
            frame_kernels = image_frame_kernels[image_index]
            frame_length = image_frame_lengths[image_index]
            value = image_array[image_index]

            for kernel_index in range(frame_length):

                vector_index = frame_indexes[kernel_index]
                kernel = frame_kernels[kernel_index]
                new_array[vector_index] += value * kernel

        for image_index in range(len(blurring_array)):

            frame_indexes = blurring_frame_indexes[image_index]
            frame_kernels = blurring_frame_kernels[image_index]
            frame_length = blurring_frame_lengths[image_index]
            value = blurring_array[image_index]

            for kernel_index in range(frame_length):

                vector_index = frame_indexes[kernel_index]
                kernel = frame_kernels[kernel_index]
                new_array[vector_index] += value * kernel

        return new_array


class ConvolverMappingMatrix(Convolver):

    def __init__(self, mask, psf):
        """
        Class to create number array and frames used to convolve a psf with a 1D vector of non-masked values.

        Parameters
        ----------
        mask : Mask
            A mask where True eliminates data.
        mask : Mask
            A mask of pixels outside the mask but whose light blurs into it after convolution.
        psf : image.PSF or ndarray
            An array representing a PSF psf.

        Attributes
        ----------
        blurring_frame_indexes: [ndarray]
            An array of frames created by the frame maker. Maps positions in the psf to values in the 1D vector for
            masked pixels.
        image_frame_indexes: [ndarray]
            An array of frames created by the frame maker. A frame maps positions in the psf to values in the 1D
            vector.
        """

        super(ConvolverMappingMatrix, self).__init__(mask, psf)

    def convolve_mapping_matrix_jit(self, mapping):
        """
        Simple version of function that applies this convolver_image to a whole mapping matrix.

        Parameters
        ----------
        blurring_array: [Float]
            An array representing the mapping of a source pixel to a set of image pixels within the blurring region.
        array: [float]
            An array representing the mapping of a source pixel to a set of image pixels.

        Returns
        -------
        convolved_array: [float]
            A matrix representing the mapping of source data_to_image to image_grid data_to_image accounting for
            convolution
        """
        return self.convolve_matrix_jitted(mapping, self.image_frame_indexes,
                                           self.image_frame_psfs, self.image_frame_lengths)

    @staticmethod
    @numba.jit(nopython=True)
    def convolve_matrix_jitted(mapping_matrix, image_frame_indexes, image_frame_kernels, image_frame_lengths):

        blurred_mapping = np.zeros(mapping_matrix.shape)

        for pixel_index in range(mapping_matrix.shape[1]):
            for image_index in range(mapping_matrix.shape[0]):

                value = mapping_matrix[image_index, pixel_index]

                if value > 0:

                    frame_indexes = image_frame_indexes[image_index]
                    frame_kernels = image_frame_kernels[image_index]
                    frame_length = image_frame_lengths[image_index]

                    for kernel_index in range(frame_length):

                        vector_index = frame_indexes[kernel_index]
                        kernel = frame_kernels[kernel_index]
                        blurred_mapping[vector_index, pixel_index] += value * kernel

        return blurred_mapping


import math
from scipy.special import erfinv
import inspect
import os
from autolens import exc
from autolens.config import DefaultPriorConfig, WidthConfig

path = os.path.dirname(os.path.realpath(__file__))


class ModelMapper(object):
    """A collection of priors formed by passing in classes to be reconstructed
        @DynamicAttrs
    """

    def __init__(self, config=None, width_config=None, **classes):
        """
        Parameters
        ----------
        config: auto_lens.config.config.Config
            An object that wraps a configuration

        Examples
        --------
        # The ModelMapper converts a set of classes whose input attributes may be modeled using a non-linear search, to
        # parameters with priors attached.

        # A config is passed into the collection to provide default setup values for the priors:

        collection = ModelMapper(config)

        # All class instances that are to be generated by the collection are specified by passing their name and class:

        collection.add_class("sersic_1", light_profile.EllipticalSersicMass)
        collection.add_class("sersic_2", light_profile.EllipticalSersicMass)
        collection.add_class("other_instance", SomeClass)

        # A PriorModel instance is created each time we add a class to the collection. We can access those models using
        # their name:

        sersic_model_1 = collection.sersic_1

        # This allows us to replace the default priors:

        collection.sersic_1.intensity = GaussianPrior(2., 5.)

        # Or maybe we want to tie two priors together:

        collection.sersic_1.intensity = collection.sersic_2.intensity

        # This statement reduces the number of priors by one and means that the two sersic instances will always share
        # the same centre.

        # We can then create instances of every class for a unit hypercube vector with length equal to
        # len(collection.priors):

        model_instance = collection.model_instance_for_vector([.4, .2, .3, .1])

        # The attributes of the model_instance are named the same as those of the collection:

        sersic_1 = collection.sersic_1

        # But this attribute is an instance of the actual EllipticalSersicMass class

        # A ModelMapper can be concisely constructed using keyword arguments:

        collection = prior.ModelMapper(config, source_light_profile=light_profile.EllipticalSersicMass,
                                    lens_mass_profile=mass_profile.EllipticalCoredIsothermal,
                                    lens_light_profile=light_profile.EllipticalCoreSersic)
        """
        super(ModelMapper, self).__init__()

        self.config = (config if config is not None else DefaultPriorConfig("{}/../../config/priors/default".format(path)))
        self.width_config = (
            width_config if width_config is not None else WidthConfig("{}/../../config/priors/width".format(path)))

        for name, cls in classes.items():
            self.__setattr__(name, cls)

    @property
    def total_parameters(self):
        return len(self.priors_ordered_by_id)

    def __setattr__(self, key, value):
        if isinstance(value, list) and len(value) > 0 and isinstance(value[0], AbstractPriorModel):
            value = ListPriorModel(value)
        elif inspect.isclass(value):
            value = PriorModel(value, config=self.config)
        super(ModelMapper, self).__setattr__(key, value)

    def add_classes(self, **kwargs):
        for key, value in kwargs.items():
            self.__setattr__(key, value)

    @property
    def prior_models(self):
        """
        Returns
        -------
        prior_model_tuples: [(String, PriorModel)]
        """
        return list(filter(lambda t: isinstance(t[1], AbstractPriorModel), self.__dict__.items()))

    @property
    def list_prior_models(self):
        """
        Returns
        -------
        list_prior_model_tuples: [(String, ListPriorModel)]
        """
        return list(filter(lambda t: isinstance(t[1], ListPriorModel), self.__dict__.items()))

    @property
    def flat_prior_models(self):
        """
        Returns
        -------
        prior_model_tuples: [(String, PriorModel)]
            A list of tuples with the names of prior models and associated prior models. Names are fully qualified by
            all objects in which they are embedded.
        """
        return [("{}".format(prior_model_name), flat_prior_model) for
                prior_model_name, prior_model in
                self.prior_models for
                flat_prior_model_name, flat_prior_model in
                prior_model.flat_prior_models]

    @property
    def prior_set(self):
        """
        Returns
        -------
        prior_set: set()
            The set of all priors associated with this collection
        """
        return {prior[1]: prior for name, prior_model in self.prior_models for prior in
                prior_model.priors}.values()

    @property
    def prior_class_dict(self):
        return {prior: cls for prior_model in self.prior_models for prior, cls in
                prior_model[1].prior_class_dict.items()}

    @property
    def priors_ordered_by_id(self):
        """
        Returns
        -------
        priors: [Prior]
            An ordered list of unique priors associated with this collection
        """
        return sorted(list(self.prior_set), key=lambda prior: prior[1].id)

    @property
    def class_priors_dict(self):
        """
        Returns
        -------
        class_priors_dict: {String: [Prior]}
            A dictionary mapping the names of reconstructable class instances to lists of associated priors
        """
        return {name: list(prior_model.priors) for name, prior_model in self.prior_models}

    def physical_vector_from_hypercube_vector(self, hypercube_vector):
        """
        Parameters
        ----------
        hypercube_vector: [float]
            A unit hypercube vector

        Returns
        -------
        values: [float]
            A vector with values output by priors
        """
        return list(map(lambda prior, unit: prior[1].value_for(unit), self.priors_ordered_by_id, hypercube_vector))

    def physical_values_ordered_by_class(self, hypercube_vector):
        model_instance = self.instance_from_unit_vector(hypercube_vector)
        result = []
        for instance_key in sorted(model_instance.__dict__.keys()):
            instance = model_instance.__dict__[instance_key]
            for attribute_key in sorted(instance.__dict__.keys()):

                value = instance.__dict__[attribute_key]

                if isinstance(value, tuple):
                    result.extend(list(value))
                else:
                    result.append(value)
        return result

    @property
    def physical_vector_from_prior_medians(self):
        return self.physical_vector_from_hypercube_vector([0.5] * len(self.prior_set))

    def instance_from_prior_medians(self):
        """
        Creates a list of physical values from the median values of the priors.

        Returns
        -------
        physical_values : [float]
            A list of physical values

        """
        return self.instance_from_unit_vector(unit_vector=[0.5] * len(self.prior_set))

    def instance_from_unit_vector(self, unit_vector):
        """
        Creates a ModelInstance, which has an attribute and class instance corresponding to every PriorModel attributed
        to this instance.

        This method takes as input a unit vector of parameter values, converting each to physical values via their \
        priors.

        Parameters
        ----------
        unit_vector: [float]
            A vector of physical parameter values.

        Returns
        -------
        model_instance : ModelInstance
            An object containing reconstructed model_mapper instances

        """
        arguments = dict(
            map(lambda prior, unit: (prior[1], prior[1].value_for(unit)), self.priors_ordered_by_id, unit_vector))

        return self.instance_from_arguments(arguments)

    def instance_from_physical_vector(self, physical_vector):
        """
        Creates a ModelInstance, which has an attribute and class instance corresponding to every PriorModel attributed
        to this instance.

        This method takes as input a physical vector of parameter values, thus omitting the use of priors.

        Parameters
        ----------
        physical_vector: [float]
            A unit hypercube vector

        Returns
        -------
        model_instance : ModelInstance
            An object containing reconstructed model_mapper instances

        """
        arguments = dict(
            map(lambda prior, physical_unit: (prior[1], physical_unit), self.priors_ordered_by_id, physical_vector))

        return self.instance_from_arguments(arguments)

    def mapper_from_prior_arguments(self, arguments):
        """
        Creates a new model mapper from a dictionary mapping existing priors to new priors.

        Parameters
        ----------
        arguments: {Prior: Prior}
            A dictionary mapping priors to priors

        Returns
        -------
        model_mapper: ModelMapper
            A new model mapper with updated priors.
        """
        mapper = ModelMapper(config=self.config, width_config=self.width_config)

        for prior_model in self.prior_models:
            setattr(mapper, prior_model[0], prior_model[1].gaussian_prior_model_for_arguments(arguments))

        return mapper

    def mapper_from_gaussian_tuples(self, tuples):
        """
        Creates a new model mapper from a list of tuples describing the mean and width values of gaussian priors. The
        new gaussian priors must be provided in the same order as the priors associated with model.

        Parameters
        ----------
        tuples: [(float, float)]
            A list of tuples containing the mean and sigma of the gaussian priors.

        Returns
        -------
        mapper: ModelMapper
            A new model mapper with all priors replaced by gaussian priors.
        """
        new_priors = map(lambda t: GaussianPrior(t[0], t[1]), tuples)
        arguments = dict(map(lambda prior, new_prior: (prior[1], new_prior), self.priors_ordered_by_id, new_priors))

        return self.mapper_from_prior_arguments(arguments)

    def mapper_from_gaussian_means(self, means):
        """
        Creates a new model mapper from a list of floats describing the mean values of gaussian priors. The widths of
        the new priors are taken from the width_config. The new gaussian priors must be provided in the same order as
        the priors associated with model.

        Parameters
        ----------
        means: [float]
            A list containing the means of the gaussian priors.

        Returns
        -------
        mapper: ModelMapper
            A new model mapper with all priors replaced by gaussian priors.
        """
        priors = self.priors_ordered_by_id
        prior_class_dict = self.prior_class_dict
        arguments = {}

        for i, prior in enumerate(priors):
            cls = prior_class_dict[prior[1]]
            width = self.width_config.get_for_nearest_ancestor(cls, prior[0])
            arguments[prior[1]] = GaussianPrior(means[i], width)

        return self.mapper_from_prior_arguments(arguments)

    def instance_from_arguments(self, arguments):
        """
        Creates a ModelInstance, which has an attribute and class instance corresponding to every PriorModel attributed
        to this instance.

        Parameters
        ----------
        arguments : dict
            The dictionary representation of prior and parameter values. This is created in the model_instance_from_* \
            routines.

        Returns
        -------
        model_instance : ModelInstance
            An object containing reconstructed model_mapper instances

        """

        model_instance = ModelInstance()

        for prior_model in self.prior_models:
            setattr(model_instance, prior_model[0], prior_model[1].instance_for_arguments(arguments))

        return model_instance

    def generate_model_info(self):
        """
        Use the priors that make up the model_mapper to information on each parameter of the overall model.

        This information is extracted from each priors *model_info* property.
        """

        model_info = ''

        for prior_name, prior_model in self.flat_prior_models:

            model_info += prior_model.cls.__name__ + '\n' + '\n'

            for i, prior in enumerate(prior_model.priors):
                param_name = str(self.class_priors_dict[prior_name][i][0])
                model_info += param_name + ': ' + (prior[1].model_info + '\n')

            model_info += '\n'

        return model_info

    def output_model_info(self, filename):
        """Output a model information file, which lists the information of the model mapper (e.g. parameters, priors,
         etc.) """
        model_info = self.generate_model_info()
        if not os.path.isfile(filename):
            with open(filename, 'w') as file:
                file.write(model_info)
            file.close()

    def check_model_info(self, filename):
        """Check whether the priors in this instance of the model_mapper are identical to those output into a model \
        info file on the hard-disk (e.g. from a previous non-linear search)."""

        model_info = self.generate_model_info()

        model_info_check = open(filename, 'r')

        if str(model_info_check.read()) != model_info:
            raise exc.PriorException(
                'The model_mapper input to MultiNest has a different prior for a parameter than the model_mapper '
                'existing in the nlo. filename = {}'.format(filename))

        model_info_check.close()


class ModelInstance(object):
    """
    An object to hold model instances produced by providing arguments to a model mapper.

    @DynamicAttrs
    """

    def instances_of(self, cls):
        return [instance for source in
                [list(self.__dict__.values())] + [ls for ls in self.__dict__.values() if isinstance(ls, list)] for
                instance in
                source if isinstance(instance, cls)]


class Constant(object):
    def __init__(self, value):
        """
        Represents a constant value. No prior is added to the model mapper for constants reducing the dimensionality
        of the nonlinear search.

        Parameters
        ----------
        value: Union(float, tuple)
            The value this constant should take.
        """
        self.value = value

    def __eq__(self, other):
        return self.value == other

    def __gt__(self, other):
        return self.value > other

    def __lt__(self, other):
        return self.value < other

    def __ne__(self, other):
        return self.value != other

    def __str__(self):
        return "Constant {}".format(self.value)


prior_number = 0


class Prior(object):
    """An object used to map a unit value to an attribute value for a specific class attribute"""

    def __init__(self):
        global prior_number
        self.id = prior_number
        prior_number += 1

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        return "<Prior id={}>".format(self.id)


class UniformPrior(Prior):
    """A prior with a uniform distribution between a lower and upper limit"""

    def __init__(self, lower_limit=0., upper_limit=1.):
        """

        Parameters
        ----------
        lower_limit: Float
            The lowest value this prior can return
        upper_limit: Float
            The highest value this prior can return
        """
        super(UniformPrior, self).__init__()
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit

    def value_for(self, unit):
        """

        Parameters
        ----------
        unit: Float
            A unit hypercube value between 0 and 1
        Returns
        -------
        value: Float
            A value for the attribute between the upper and lower limits
        """
        return self.lower_limit + unit * (self.upper_limit - self.lower_limit)

    @property
    def model_info(self):
        """The line of text describing this prior for the model_mapper.info file"""
        return 'UniformPrior, lower_limit = ' + str(self.lower_limit) + ', upper_limit = ' + str(self.upper_limit)


class GaussianPrior(Prior):
    """A prior with a gaussian distribution"""

    def __init__(self, mean, sigma):
        super(GaussianPrior, self).__init__()
        self.mean = mean
        self.sigma = sigma

    def value_for(self, unit):
        """

        Parameters
        ----------
        unit: Float
            A unit hypercube value between 0 and 1
        Returns
        -------
        value: Float
            A value for the attribute biased to the gaussian distribution
        """
        return self.mean + (self.sigma * math.sqrt(2) * erfinv((unit * 2.0) - 1.0))

    @property
    def model_info(self):
        """The line of text describing this prior for the model_mapper.info file"""
        return 'GaussianPrior, mean = ' + str(self.mean) + ', sigma = ' + str(self.sigma)


class TuplePrior(object):
    """
    A prior comprising one or more priors in a tuple
    """

    @property
    def priors(self):
        """
        Returns
        -------
        priors: [(String, Prior)]
            A list of priors contained in this tuple
        """
        return list(filter(lambda t: isinstance(t[1], Prior), self.__dict__.items()))

    def value_for_arguments(self, arguments):
        """
        Parameters
        ----------
        arguments: {Prior: float}
            A dictionary of arguments

        Returns
        -------
        tuple: (float,...)
            A tuple of float values
        """
        return tuple([arguments[prior[1]] for prior in self.priors])

    def gaussian_tuple_prior_for_arguments(self, arguments):
        """
        Parameters
        ----------
        arguments: {Prior: float}
            A dictionary of arguments

        Returns
        -------
        tuple_prior: TuplePrior
            A new tuple prior with gaussian priors
        """
        tuple_prior = TuplePrior()
        for prior in self.priors:
            setattr(tuple_prior, prior[0], arguments[prior[1]])
        return tuple_prior


class AbstractPriorModel:
    """
    Abstract model that maps a set of priors to a particular class. Must be overridden by any prior model so that the
    model mapper recognises its prior model attributes.
    """

    @property
    def flat_prior_models(self):
        """
        Returns
        -------
        prior_models: [(str, AbstractPriorModel)]
            A list of prior models associated with this instance
        """
        raise NotImplementedError("PriorModels must implement the flat_prior_models property")


class PriorModel(AbstractPriorModel):
    """Object comprising class and associated priors
        @DynamicAttrs
    """

    @property
    def flat_prior_models(self):
        return [("", self)]

    def __init__(self, cls, config=None):
        """
        Parameters
        ----------
        cls: class
            The class associated with this instance
        """

        self.cls = cls
        self.config = (config if config is not None else DefaultPriorConfig("{}/../../config/priors/default".format(path)))

        arg_spec = inspect.getfullargspec(cls.__init__)

        try:
            defaults = dict(zip(arg_spec.args[-len(arg_spec.defaults):], arg_spec.defaults))
        except TypeError:
            defaults = {}

        args = arg_spec.args[1:]

        if 'settings' in defaults:
            del defaults['settings']
        if 'settings' in args:
            args.remove('settings')

        for arg in args:
            if arg in defaults and isinstance(defaults[arg], tuple):
                tuple_prior = TuplePrior()
                for i in range(len(defaults[arg])):
                    attribute_name = "{}_{}".format(arg, i)
                    setattr(tuple_prior, attribute_name, self.make_prior(attribute_name, cls))
                setattr(self, arg, tuple_prior)
            else:
                setattr(self, arg, self.make_prior(arg, cls))

    def make_prior(self, attribute_name, cls):
        config_arr = self.config.get_for_nearest_ancestor(cls, attribute_name)
        if config_arr[0] == "u":
            return UniformPrior(config_arr[1], config_arr[2])
        elif config_arr[0] == "g":
            return GaussianPrior(config_arr[1], config_arr[2])
        elif config_arr[0] == "c":
            return Constant(config_arr[1])

    @property
    def tuple_priors(self):
        """
        Returns
        -------
        tuple_prior_tuples: [(String, TuplePrior)]
        """
        return list(filter(lambda t: isinstance(t[1], TuplePrior), self.__dict__.items()))

    @property
    def direct_priors(self):
        """
        Returns
        -------
        direct_priors: [(String, Prior)]
        """
        return list(filter(lambda t: isinstance(t[1], Prior), self.__dict__.items()))

    @property
    def priors(self):
        """
        Returns
        -------
        priors: [(String, Union(Prior, TuplePrior))]
        """
        return [prior for tuple_prior in self.tuple_priors for prior in
                tuple_prior[1].priors] + self.direct_priors

    @property
    def constants(self):
        """
        Returns
        -------
        constants: [(String, Constant)]
            A list of constants
        """
        return list(filter(lambda t: isinstance(t[1], Constant), self.__dict__.items()))

    @property
    def prior_class_dict(self):
        return {prior[1]: self.cls for prior in self.priors}

    def instance_for_arguments(self, arguments):
        """
        Create an instance of the associated class for a set of arguments

        Parameters
        ----------
        arguments: {Prior: value}
            Dictionary mapping priors to attribute name and value pairs

        Returns
        -------
            An instance of the class
        """
        model_arguments = {t[0]: arguments[t[1]] for t in self.direct_priors}
        for tuple_prior in self.tuple_priors:
            model_arguments[tuple_prior[0]] = tuple_prior[1].value_for_arguments(arguments)

        constant_arguments = {t[0]: t[1].value for t in self.constants}

        return self.cls(**{**model_arguments, **constant_arguments})

    def gaussian_prior_model_for_arguments(self, arguments):
        """
        Create a new instance of model mapper with a set of Gaussian priors based on tuples provided by a previous
        nonlinear search.

        Parameters
        ----------
        arguments: [(float, float)]
            Tuples providing the mean and sigma of gaussians

        Returns
        -------
        new_model: ModelMapper
            A new model mapper populated with Gaussian priors
        """
        new_model = PriorModel(self.cls, self.config)

        model_arguments = {t[0]: arguments[t[1]] for t in self.direct_priors}

        for tuple_prior in self.tuple_priors:
            setattr(new_model, tuple_prior[0], tuple_prior[1].gaussian_tuple_prior_for_arguments(arguments))
        for prior in self.direct_priors:
            setattr(new_model, prior[0], model_arguments[prior[0]])
        for constant in self.constants:
            setattr(new_model, constant[0], constant[1])

        return new_model


class ListPriorModel(list, AbstractPriorModel):
    @property
    def flat_prior_models(self):
        return [flat_prior_model for prior_model in self for flat_prior_model in prior_model.flat_prior_models]

    def __init__(self, prior_models):
        """
        A prior model used to represent a list of prior models for convenience.

        Parameters
        ----------
        prior_models: [PriorModel]
            A list of prior models
        """
        super().__init__(prior_models)

    def instance_for_arguments(self, arguments):
        """
        Parameters
        ----------
        arguments: {Prior: float}
            A dictionary of arguments

        Returns
        -------
        model_instances: [object]
            A list of instances constructed from the list of prior models.
        """
        return [prior_model.instance_for_arguments(arguments) for prior_model in self]

    def gaussian_prior_model_for_arguments(self, arguments):
        """
        Parameters
        ----------
        arguments: {Prior: float}
            A dictionary of arguments

        Returns
        -------
        prior_models: [PriorModel]
            A new list of prior models with gaussian priors
        """
        return ListPriorModel(
            [prior_model.gaussian_prior_model_for_arguments(arguments) for prior_model in self])

    @property
    def priors(self):
        """
        Returns
        -------
        priors: [(String, Union(Prior, TuplePrior))]
        """
        return set([prior for prior_model in self for prior in prior_model.priors])

    @property
    def prior_class_dict(self):
        return {prior: cls for prior_model in self for prior, cls in prior_model.prior_class_dict.items()}



import sys
sys.path.append("../")

from autolens.analysis import fitting

import numpy as np

# Image 1 - lets pretend units are electrons per second

image = np.array([1.0])
noise = np.array([0.1])
model = np.array([0.5])

chi_sq_0 = fitting.compute_chi_sq_term(image=image, noise=noise, model_image=model)
noise_term_0 = fitting.compute_noise_term(noise)
likelihood_0 = fitting.compute_likelihood(image=image, noise=noise, model_image=model)

# Image 2 - Same image_coords but electron counts (x 10000)

image = np.array([10000.0])
noise = np.array([1000.0])
model = np.array([5000.0])

chi_sq_1 = fitting.compute_chi_sq_term(image=image, noise=noise, model_image=model)
noise_term_1 = fitting.compute_noise_term(noise)
likelihood_1 = fitting.compute_likelihood(image=image, noise=noise, model_image=model)

print('The likelihoods are different, due to the normalization terms')
print('(However, we only care about potential changes in likelihood)')
print()

print(chi_sq_0, chi_sq_1)
print(noise_term_0, noise_term_1)
print(likelihood_0, likelihood_1)


# Lets scale the noise in both images by x3, does the change in likelihood care about the image_coords units?

image = np.array([1.0])
noise = np.array([0.3])
model = np.array([0.5])

chi_sq_2 = fitting.compute_chi_sq_term(image=image, noise=noise, model_image=model)
noise_term_2 = fitting.compute_noise_term(noise)
likelihood_2 = fitting.compute_likelihood(image=image, noise=noise, model_image=model)

image = np.array([10000.0])
noise = np.array([3000.0])
model = np.array([5000.0])

chi_sq_3 = fitting.compute_chi_sq_term(image=image, noise=noise, model_image=model)
noise_term_3 = fitting.compute_noise_term(noise)
likelihood_3 = fitting.compute_likelihood(image=image, noise=noise, model_image=model)

print()
print('Likelihood normalizations different again')
print()
print(chi_sq_2)
print(noise_term_2)
print(chi_sq_2 + noise_term_2)

print(chi_sq_3)
print(noise_term_3)
print(chi_sq_3 + noise_term_3)

print()

print('Noise terms are on different scales...')
print(noise_term_1 - noise_term_0)
print(noise_term_3 - noise_term_2)
print(noise_term_0 - noise_term_2)
print(noise_term_1 - noise_term_3)
print()
print('But, magically, the change in likelihoods are identical (i.e. noise units dont impact likelihood scale)')
print((likelihood_0) - (likelihood_1))
print((likelihood_2) - (likelihood_3))
print((likelihood_0) - (likelihood_2))
print((likelihood_1) - (likelihood_3))


"""
skele

Usage:
  skele hello
  skele -h | --help
  skele --version

Options:
  -h --help                         Show this screen.
  --version                         Show version.

Examples:
  skele hello

Help:
  For help using this tool, please open an issue on the Github repository:
  https://github.com/rdegges/skele-cli
"""

from inspect import getmembers, isclass

from docopt import docopt

from . import __version__ as VERSION


def main():
    """Main CLI entrypoint."""
    import autolens.commands
    options = docopt(__doc__, version=VERSION)

    # Here we'll try to dynamically match the command the user is trying to run
    # with a pre-defined command class we've already created.
    for (k, v) in options.items():
        if hasattr(autolens.commands, k) and v:
            module = getattr(autolens.commands, k)
            command = [command[1] for command in getmembers(module, isclass) if command[0] != 'Base'][0]
            command = command(options)
            command.run()



from autolens.analysis import galaxy
import inspect
from autolens.profiles import light_profiles, mass_profiles
from autolens.autopipe import model_mapper


def is_light_profile_class(cls):
    return inspect.isclass(cls) and issubclass(
        cls, light_profiles.LightProfile) and not issubclass(
        cls, mass_profiles.MassProfile)


def is_mass_profile_class(cls):
    return inspect.isclass(cls) and issubclass(cls, mass_profiles.MassProfile)


def is_profile_class(cls):
    return inspect.isclass(cls) \
           and issubclass(cls, mass_profiles.MassProfile) \
           or issubclass(cls, light_profiles.LightProfile)


class GalaxyPrior(model_mapper.AbstractPriorModel):
    """
    @DynamicAttrs
    """

    @property
    def flat_prior_models(self):
        return [flat_prior_model for prior_model in self.prior_models for flat_prior_model in
                prior_model.flat_prior_models]

    def __init__(self, align_centres=False, align_orientations=False, redshift=None, variable_redshift=False,
                 pixelization=None, hyper_galaxy=None, config=None, **kwargs):
        """
        Class to produce Galaxy instances from sets of profile classes using the model mapper

        Parameters
        ----------
        light_profile_classes: [LightProfile]
            The classes for which light profile instances are generated for this galaxy
        mass_profile_classes: [MassProfile]
            The classes for which light profile instances are generated for this galaxy
        align_centres: Bool
            If True the same prior will be used for all the profiles centres such that any generated profiles always
            have the same centre
        align_orientations: Bool
            If True the same prior will be used for all the profiles orientations such that any generated profiles
            always have the same orientation
        """

        self.align_centres = align_centres
        self.align_orientations = align_orientations

        profile_models = []

        for name, cls in kwargs.items():
            if is_mass_profile_class(cls) or is_light_profile_class(cls):
                model = model_mapper.PriorModel(cls, config)
                profile_models.append(model)
                setattr(self, name, model)
            else:
                setattr(self, name, cls)

        if len(profile_models) > 0:
            if self.align_centres:
                centre = profile_models[0].centre
                for profile_model in profile_models:
                    profile_model.centre = centre

            if self.align_orientations:
                phi = profile_models[0].phi
                for profile_model in profile_models:
                    profile_model.phi = phi

        if redshift is not None:
            self.redshift = model_mapper.Constant(
                redshift.redshift if isinstance(redshift, galaxy.Redshift) else redshift)
        else:
            self.redshift = model_mapper.PriorModel(galaxy.Redshift,
                                                    config) if variable_redshift else model_mapper.Constant(1)

        self.pixelization = model_mapper.PriorModel(pixelization, config) if inspect.isclass(
            pixelization) else pixelization
        self.hyper_galaxy = model_mapper.PriorModel(hyper_galaxy, config) if inspect.isclass(
            hyper_galaxy) else hyper_galaxy
        self.config = config

    def __setattr__(self, key, value):
        if key == "redshift" \
                and (isinstance(value, float)
                     or isinstance(value, int)
                     or isinstance(value, model_mapper.Prior)
                     or isinstance(value, model_mapper.Constant)):
            value = galaxy.Redshift(value)
        super(GalaxyPrior, self).__setattr__(key, value)

    @property
    def constant_light_profiles(self):
        """
        Returns
        -------
        light_profiles: [light_profiles.LightProfile]
            Light profiles with set variables
        """
        return [value for value in self.__dict__.values() if galaxy.is_light_profile(value)]

    @property
    def constant_mass_profiles(self):
        """
        Returns
        -------
        mass_profiles: [mass_profiles.MassProfile]
            Mass profiles with set variables
        """
        return [value for value in self.__dict__.values() if galaxy.is_mass_profile(value)]

    @property
    def prior_models(self):
        """
        Returns
        -------
        prior_models: [model_mapper.PriorModel]
            A list of the prior models (e.g. variable profiles) attached to this galaxy prior
        """
        return [value for _, value in
                filter(lambda t: isinstance(t[1], model_mapper.PriorModel), self.__dict__.items())]

    @property
    def profile_prior_model_dict(self):
        """
        Returns
        -------
        profile_prior_model_dict: {str: PriorModel}
            A dictionary mapping instance variable names to variable profiles.
        """
        return {key: value for key, value in
                filter(lambda t: isinstance(t[1], model_mapper.PriorModel) and is_profile_class(t[1].cls),
                       self.__dict__.items())}

    @property
    def constant_profile_dict(self):
        """
        Returns
        -------
        constant_profile_dict: {str: geometry_profiles.Profile}
            A dictionary mapping instance variable names to profiles with set variables.
        """
        return {key: value for key, value in self.__dict__.items() if
                galaxy.is_light_profile(value) or galaxy.is_mass_profile(value)}

    @property
    def light_profile_prior_model_dict(self):
        """
        Returns
        -------
        profile_prior_model_dict: {str: PriorModel}
            A dictionary mapping instance variable names to variable light profiles.
        """
        return {key: value for key, value in self.prior_model_dict.items() if is_light_profile_class(value.cls)}

    @property
    def mass_profile_prior_model_dict(self):
        """
        Returns
        -------
        profile_prior_model_dict: {str: PriorModel}
            A dictionary mapping instance variable names to variable mass profiles.
        """
        return {key: value for key, value in self.prior_model_dict.items() if is_mass_profile_class(value.cls)}

    @property
    def priors(self):
        """
        Returns
        -------
        priors: [Prior]
            A list of priors associated with prior models in this galaxy prior.
        """
        return [prior for prior_model in self.prior_models for prior in prior_model.priors]

    @property
    def prior_class_dict(self):
        """
        Returns
        -------
        prior_class_dict: {Prior: class}
            A dictionary mapping priors to the class associated with their prior model.
        """
        return {prior: cls for prior_model in self.prior_models for prior, cls in
                prior_model.prior_class_dict.items()}

    def instance_for_arguments(self, arguments):
        """
        Create an instance of the associated class for a set of arguments

        Parameters
        ----------
        arguments: {Prior: value}
            Dictionary mapping priors to attribute name and value pairs

        Returns
        -------
            An instance of the class
        """
        profiles = {**{key: value.instance_for_arguments(arguments)
                       for key, value
                       in self.profile_prior_model_dict.items()}, **self.constant_profile_dict}

        if isinstance(self.redshift, galaxy.Redshift):
            redshift = self.redshift
        else:
            redshift = self.redshift.instance_for_arguments(arguments)
        pixelization = self.pixelization.instance_for_arguments(arguments) \
            if isinstance(self.pixelization, model_mapper.PriorModel) \
            else self.pixelization
        hyper_galaxy = self.hyper_galaxy.instance_for_arguments(arguments) \
            if isinstance(self.hyper_galaxy, model_mapper.PriorModel) \
            else self.hyper_galaxy

        return galaxy.Galaxy(redshift=redshift.redshift, pixelization=pixelization, hyper_galaxy=hyper_galaxy,
                             **profiles)

    def gaussian_prior_model_for_arguments(self, arguments):
        """
        Create a new galaxy prior from a set of arguments, replacing the priors of some of this galaxy prior's prior
        models with new arguments.

        Parameters
        ----------
        arguments: dict
            A dictionary mapping between old priors and their replacements.

        Returns
        -------
        new_model: GalaxyPrior
            A model with some or all priors replaced.
        """
        new_model = GalaxyPrior(align_centres=self.align_centres, align_orientations=self.align_orientations,
                                config=self.config)

        for key, value in filter(lambda t: isinstance(t[1], model_mapper.PriorModel), self.__dict__.items()):
            setattr(new_model, key, value.gaussian_prior_model_for_arguments(arguments))

        return new_model

    @classmethod
    def from_galaxy(cls, g, **kwargs):
        """
        Create a new galaxy prior with constants taken from a galaxy.

        Parameters
        ----------
        g: galaxy.Galaxy
            A galaxy
        kwargs
            Key word arguments to override GalaxyPrior constructor arguments.
        """
        return GalaxyPrior(**{**g.__dict__, **kwargs})
